/*###################################################################################
#
#   Embperl - Copyright (c) 1997-2000 Gerald Richter / ECOS
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.
#
#   THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#   WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
###################################################################################*/


#include "ep.h"
#include "epmacro.h"

struct tEmbperlCmd
    {
    const char * *  sPerlCode ;
    const char * *  sCompileTimePerlCode ;
    const char *    sPerlCodeEnd ;
    const char *    sStackName ;
    const char *    sPushStack ;
    const char *    sMatchStack ;
    const char *    sStackName2 ;
    const char *    sPushStack2 ;
    int		    numPerlCode ;
    int		    numCompileTimePerlCode ;
    int		    bRemoveNode ;
    int		    bPerlCodeRemove ;
    int		    bCompileChilds ;
    int		    nNodeType ;
    int		    nSwitchCodeType ;
    const char *    sMayJump ;
    } ;

typedef struct tEmbperlCmd tEmbperlCmd ;

char * * pProg = NULL ;
char * pProgRun = NULL ;
char * pProgDef = NULL ;

int nLabelCnt ;

tStringIndex nMaxEmbperlCmd ;
struct tEmbperlCmd * pEmbperlCmds ;


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileInit                                                      */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int embperl_CompileInit (void)

    {
    ArrayNew (&pEmbperlCmds, 256, sizeof (struct tEmbperlCmd)) ; 
    ArraySet (&pEmbperlCmds, 0) ;
    nMaxEmbperlCmd = 1 ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileInitItem                                                  */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int embperl_CompileInitItem      (/*i/o*/ register req * r,
				  /*in*/  HV *           pHash,
				  /*in*/  int            nNodeName,
				  /*in*/  int            nNodeType)

    {
    SV * * ppSV ;
    AV * pAV ;
    
    
    ArraySet (&pEmbperlCmds, nNodeName) ;
    if (nMaxEmbperlCmd < nNodeName) 
	nMaxEmbperlCmd = nNodeName ;

    ppSV = hv_fetch(pHash, "perlcode", 8, 0) ;  
    if (ppSV != NULL && *ppSV != NULL && 
        SvTYPE(*ppSV) == SVt_RV && SvTYPE((pAV = (AV *)SvRV(*ppSV))) == SVt_PVAV)
	{ /* Array reference  */
	int f = AvFILL(pAV) + 1 ;
        int i ;
        STRLEN l ;
        char * s ;

        pEmbperlCmds[nNodeName].sPerlCode = malloc (f * sizeof (char *)) ;
        pEmbperlCmds[nNodeName].numPerlCode = f ;

        for (i = 0; i < f; i++)
	    {
	    ppSV = av_fetch (pAV, i, 0) ;
	    if (ppSV && *ppSV)
		pEmbperlCmds[nNodeName].sPerlCode[i] = strdup (SvPV (*ppSV, l)) ;
            else
		pEmbperlCmds[nNodeName].sPerlCode[i] = NULL ;
            }
        }
    else
        {
        if (ppSV)
            {
            STRLEN  l ; 
            
            pEmbperlCmds[nNodeName].sPerlCode = malloc (sizeof (char *)) ;
            pEmbperlCmds[nNodeName].numPerlCode = 1 ;
            pEmbperlCmds[nNodeName].sPerlCode[0] = sstrdup (SvPV (*ppSV, l)) ;
            }
        }
    

    ppSV = hv_fetch(pHash, "compiletimeperlcode", 19, 0) ;  
    if (ppSV != NULL && *ppSV != NULL && 
        SvTYPE(*ppSV) == SVt_RV && SvTYPE((pAV = (AV *)SvRV(*ppSV))) == SVt_PVAV)
	{ /* Array reference  */
	int f = AvFILL(pAV) + 1 ;
        int i ;
        STRLEN l ;
        char * s ;

        pEmbperlCmds[nNodeName].sCompileTimePerlCode = malloc (f * sizeof (char *)) ;
        pEmbperlCmds[nNodeName].numCompileTimePerlCode = f ;

        for (i = 0; i < f; i++)
	    {
	    ppSV = av_fetch (pAV, i, 0) ;
	    if (ppSV && *ppSV)
		pEmbperlCmds[nNodeName].sCompileTimePerlCode[i] = strdup (SvPV (*ppSV, l)) ;
            else
		pEmbperlCmds[nNodeName].sCompileTimePerlCode[i] = NULL ;
            }
        }
    else
        {
        if (ppSV)
            {
            STRLEN  l ; 
            
            pEmbperlCmds[nNodeName].sCompileTimePerlCode = malloc (sizeof (char *)) ;
            pEmbperlCmds[nNodeName].numCompileTimePerlCode = 1 ;
            pEmbperlCmds[nNodeName].sCompileTimePerlCode[0] = sstrdup (SvPV (*ppSV, l)) ;
            }
        }
    
    
    
    
    pEmbperlCmds[nNodeName].sPerlCodeEnd = GetHashValueStrDup  (pHash, "perlcodeend", NULL) ;
    pEmbperlCmds[nNodeName].sStackName   = GetHashValueStrDup  (pHash, "stackname", NULL) ;
    pEmbperlCmds[nNodeName].sPushStack   = GetHashValueStrDup  (pHash, "push", NULL) ;
    pEmbperlCmds[nNodeName].sMatchStack   = GetHashValueStrDup  (pHash, "stackmatch", NULL) ;
    pEmbperlCmds[nNodeName].sStackName2   = GetHashValueStrDup (pHash, "stackname2", NULL) ;
    pEmbperlCmds[nNodeName].sPushStack2   = GetHashValueStrDup (pHash, "push2", NULL) ;
    pEmbperlCmds[nNodeName].bRemoveNode  = GetHashValueInt     (pHash, "removenode", 0) ;
    pEmbperlCmds[nNodeName].sMayJump     = GetHashValueStrDup  (pHash, "mayjump", NULL) ;
    pEmbperlCmds[nNodeName].bPerlCodeRemove  = GetHashValueInt (pHash, "perlcoderemove", 0) ;
    pEmbperlCmds[nNodeName].bCompileChilds  = GetHashValueInt  (pHash, "compilechilds", 1) ;
    pEmbperlCmds[nNodeName].nSwitchCodeType  = GetHashValueInt  (pHash, "switchcodetype", 0) ;
    pEmbperlCmds[nNodeName].nNodeType  = nNodeType ;


    if (r -> bDebug & dbgParse)
        lprintf (r, "[%d]EPCOMP: InitItem %s (#%d) perlcode=%s (num=%d) perlcodeend=%s\n", 
	                  r -> nPid, Ndx2String(nNodeName), nNodeName, 
			  pEmbperlCmds[nNodeName].sPerlCode?pEmbperlCmds[nNodeName].sPerlCode[0]:"", 
			  pEmbperlCmds[nNodeName].numPerlCode, 
			  pEmbperlCmds[nNodeName].sPerlCodeEnd?pEmbperlCmds[nNodeName].sPerlCodeEnd:"<undef>") ; 

    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* strstrn			                                            */
/*                                                                          */
/* find substring of length n                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */

static const char * strstrn (const char * s1, const char * s2, int l)

    {
    while (*s1)
	{
	if ((s1 = strchr (s1, *s2)) == NULL)
	    return NULL ;
	if (strncmp (s1, s2, l) == 0)
	    return s1 ;
	s1++ ;
	}

    return NULL ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileAddValue                                                  */
/*                                                                          */
/* Add value to perl code                                                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileAddValue    (/*in*/  const char * sText,
				        const char * p,
					const char * q,
                                        const char * eq, 
					char op,
					char out,
                                       /*i/o*/  char * *      ppCode )




    {
    const char * or ;
    const char * e ;

    
    if (sText)
	{
        if (op == '=' && eq)
	    {
	    eq++ ;
	    do
		{
		or = strchr (eq + 1, '|') ;
		e = or?or:q ;
		if (strncmp (sText, eq, e - eq) == 0)
		    break ;
		if (or == NULL)
		    return 0 ;
		eq = or + 1 ;
		}
	    while (or) ;
	    }
	else if (op == '~' && eq)
	    {
	    eq++ ;
	    do 
		{
		or = strchr (eq + 1, '|') ;
		e = or?or:q ;
		if (strstrn (sText, eq, e - eq))
		    break ;
		if (or == NULL)
		    return 0 ;
		eq = or + 1 ;
		}
	    while (or) ;
	    }
	else if (op == '!' && sText)
	    {
	    return 0 ;
	    }

        if (out)
	    {
	    if (out == 2)
		{
		StringAdd (ppCode, "'", 1) ;
		StringAdd (ppCode, sText, 0) ;
		StringAdd (ppCode, "'", 1) ;
		}
	    else if (out)            
		StringAdd (ppCode, sText, 0) ;
	    }
        }
    else
        {
        if (op != '!' && op != 0)
	    return 0 ;
    	
	if (out == 2)
	    StringAdd (ppCode, "''", 2) ;
	else  if (out)           
	    StringAdd (ppCode, "undef", 5) ;
	}

    return 1 ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompilePushStack                                                 */
/*                                                                          */
/* Push valuie on named stack                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static void embperl_CompilePushStack    (/*in*/ tDomTree *   pDomTree,
				              const char * sStackName,
				              const char * sStackValue)
                                              
    {
    SV **   ppSV ;
    SV *    pSV ;
    AV *    pAV ;

    ppSV = hv_fetch((HV *)(pDomTree -> pSV), (char *)sStackName, strlen (sStackName), 1) ;  
    if (ppSV == NULL)
        return  ;

    if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_RV)
	{
	if (*ppSV)
	    SvREFCNT_dec (*ppSV) ;
        *ppSV = newRV_noinc ((SV *)(pAV = newAV ())) ;
        }
    else
        pAV = (AV *)SvRV (*ppSV) ;
        
        
    pSV = newSVpv ((char *)sStackValue, strlen (sStackValue)) ;
    SvUPGRADE (pSV, SVt_PVIV) ;
    SvIVX (pSV) = 0 ;
    av_push (pAV, pSV) ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompilePopStack                                                  */
/*                                                                          */
/* pop value from named stack                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static void embperl_CompilePopStack    (/*in*/ tDomTree *   pDomTree,
				              const char * sStackName)
                                              
    {
    SV **   ppSV ;
    SV *    pSV ;

    ppSV = hv_fetch((HV *)(pDomTree -> pSV), (char *)sStackName, strlen (sStackName), 0) ;  
    if (ppSV == NULL || *ppSV == NULL || SvTYPE (*ppSV) != SVt_RV)
        return  ;

    pSV = av_pop ((AV *)SvRV (*ppSV)) ;
    SvREFCNT_dec (pSV) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileMatchStack                                                */
/*                                                                          */
/* check if top of stack value matches given value                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileMatchStack (/*in*/ tDomTree *   pDomTree,
					     tNodeData *  pNode,
					     const char * sStackName,
				             const char * sStackValue)
                                              
    {
    SV **   ppSV ;
    SV *    pSV ;
    AV *    pAV ;
    STRLEN  l ;
    char *  s ;

    ppSV = hv_fetch((HV *)(pDomTree -> pSV), (char *)sStackName, strlen (sStackName), 0) ;  
    if (ppSV == NULL || *ppSV == NULL || SvTYPE (*ppSV) != SVt_RV)
        return rcHashError ;

    pSV = av_pop ((AV *)SvRV (*ppSV)) ;

    s = SvPV (pSV, l) ;
    if (strcmp (s, sStackValue) == 0)
	{
        SvREFCNT_dec (pSV) ;
	return ok ;
	}
	
    strncpy (pCurrReq -> errdat1, Node_selfNodeName (pNode), sizeof (pCurrReq -> errdat1)) ;
    sprintf (pCurrReq -> errdat2, "'%s', starttag should be '%s'", s, sStackValue) ;
    pCurrReq -> Buf.pCurrPos	 = NULL ;
    pCurrReq -> Buf.nSourceline = pNode -> nLinenumber ;

    SvREFCNT_dec (pSV) ;

    return rcTagMismatch ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileAddStack                                                  */
/*                                                                          */
/* Add value of child node to perl code                                     */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileAddStack    (/*in*/ tDomTree *   pDomTree,
				                const char * p,
					        const char * q,
					        char op,
					        char out,
                                                char str,
                                       /*i/o*/  char * *      ppCode )
    {
    const char * eq = strchr (p, ':') ;
    const char * e = eq && eq < q?eq:q;
    STRLEN           l ;
    const char * sText = NULL ;
    SV **   ppSV ;
    AV *    pAV ;


    ppSV = hv_fetch((HV *)(pDomTree -> pSV), (char *)p, e - p, 0) ;  
    if (ppSV == NULL || *ppSV == NULL || SvTYPE (*ppSV) != SVt_RV)
        return  op == '!'?1:0 ;

    pAV = (AV *)SvRV (*ppSV) ;

    if (SvTYPE (pAV) != SVt_PVAV)
        return  op == '!'?1:0 ;
    
    ppSV = av_fetch (pAV, av_len (pAV), 0) ;
    if (ppSV == NULL || *ppSV == NULL)
        return  op == '!'?1:0 ;

    if (str)
        {
        sText = SvPV (*ppSV, l) ;    
        (SvIVX (*ppSV))++ ;
        }
    else
        sText = SvIVX (*ppSV)?"1":NULL ;


    return embperl_CompileAddValue (sText, p, q, eq, op, out, ppCode) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileAddChildNode                                              */
/*                                                                          */
/* Add value of child node to perl code                                     */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileAddChildNode (/*in*/ tDomTree *   pDomTree,
    		                 /*in*/ tNodeData *	 pNode,
				        const char * p,
					const char * q,
					char op,
					char out,
                                       /*i/o*/  char * *      ppCode )



    {
    const char * eq = strchr (p, ':') ;
    int nChildNo = atoi (p) ;
    struct tNodeData * pChildNode = Node_selfNthChild (pDomTree, pNode, nChildNo) ;
    const char * sText = NULL ;
    
    if (pChildNode)
	sText = Node_selfNodeName(pChildNode) ;

    return embperl_CompileAddValue (sText, p, q, eq, op, out, ppCode) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileAddSiblingNode                                            */
/*                                                                          */
/* Add value of sibling node to perl code                                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileAddSiblingNode (/*in*/ tDomTree *   pDomTree,
    		                 /*in*/ tNodeData *	 pNode,
				        const char * p,
					const char * q,
					char op,
					char out,
                                       /*i/o*/  char * *      ppCode )



    {
    const char * eq = strchr (p, ':') ;
    int nChildNo = atoi (p) ;
    struct tNodeData * pChildNode = Node_selfNextSibling (pDomTree, pNode) ;
    const char * sText = NULL ;
    
    while (pChildNode && nChildNo-- > 0)
	pChildNode = Node_selfNextSibling (pDomTree, pChildNode) ;
	
    if (pChildNode)
	sText = Node_selfNodeName(pChildNode) ;

    return embperl_CompileAddValue (sText, p, q, eq, op, out, ppCode) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileAddAttribut                                               */
/*                                                                          */
/* Add value of child node to perl code                                     */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileAddAttribut (/*in*/   tDomTree *   pDomTree,
    		                       /*in*/   tNodeData *	 pNode,
				                const char * p,
					        const char * q,
					        char op,
					        char out,
                                       /*i/o*/  char * *      ppCode )



    {
    const char * eq = strchr (p, ':') ;
    const char * e = eq && eq < q?eq:q;
    tAttrData * pChildNode = Element_selfGetAttribut (pDomTree, pNode, p, e - p) ;
    const char * sText = NULL ;
    char buf [128] ;

    
    if (pChildNode)
	{
	if (pChildNode -> bFlags & aflgAttrChilds)
	    {
	    int l = sprintf (buf, "XML::Embperl::DOM::Attr_value ($_ep_DomTree,%d)", pChildNode -> xNdx) ;
	    sText = buf ;
	    if (out == 2)
		out = 1 ;
	    }
	else
	    {
	    sText = Ndx2String (pChildNode -> xValue) ;
	    }

        }
    
    return embperl_CompileAddValue (sText, p, q, eq, op, out, ppCode) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileToPerlCode                                                */
/*                                                                          */
/* Compile one command inside a node                                        */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileToPerlCode  (/*in*/ tDomTree *    pDomTree,
    		                       /*in*/ tNodeData *   pNode,
                                       /*in*/ const char *  sPerlCode,
                                       /*out*/ char * *     ppCode )


    {
    const char * p ;
    const char * q ;
    int    valid = 1 ;

    StringNew (ppCode, 512) ;
    if (sPerlCode)
	{
	p = strchr (sPerlCode, '%') ;	
	while (p)
	    {
	    int n = p - sPerlCode ;
	    if (n)
		StringAdd (ppCode, sPerlCode, n) ;
	    q = strchr (p+1, '%') ;	
	    if (q)
		{
		char  type  ;
		char  op  ;
		char  out = 1 ;

		p++ ;
		type = *p ;
		p++ ;
		op = *p ;
		if (op != '=' && op != '*' && op != '!' && op != '~')
		    op = 0 ;
		else
		    p++ ;

		if (*p == '-')
		    out = 0, p++ ;
		else if (*p == '\'')
		    out = 2, p++ ;

		
		if (type == '#')
		    {
		    if (!embperl_CompileAddChildNode (pDomTree, pNode ,p, q, op, out, ppCode))
			{
			valid = 0 ;
			break ;
			}
		    }
		else if (type == '>')
		    {
		    if (!embperl_CompileAddSiblingNode (pDomTree, pNode ,p, q, op, out, ppCode))
			{
			valid = 0 ;
			break ;
			}
		    }
		else if (type == '&')
		    {
		    if (!embperl_CompileAddAttribut (pDomTree, pNode ,p, q, op, out, ppCode))
			{
			valid = 0 ;
			break ;
			}
		    }
		else if (type == '^')
		    {
		    if (!embperl_CompileAddStack (pDomTree, p, q, op, out, 1, ppCode))
			{
			valid = 0 ;
			break ;
			}
		    }
		else if (type == '?')
		    {
		    if (!embperl_CompileAddStack (pDomTree, p, q, op, out, 0, ppCode))
			{
			valid = 0 ;
			break ;
			}
		    }
		else if (type == '$')
		    {
		    if (*p == 'n')
			{
			char s [20] ;
			int  l = sprintf (s, "$_ep_DomTree,%u", pNode -> xNdx) ;
			StringAdd (ppCode, s, l) ; 
			}
		    else if (*p == 't')
			{
			StringAdd (ppCode, "$_ep_DomTree", 0) ; 
			}
		    else if (*p == 'l')
			{
			char s [20] ;
			int  l = sprintf (s, "%u", pNode -> xNdx) ;
			StringAdd (ppCode, s, l) ; 
			}
		    }

		sPerlCode = q + 1 ;
		p = strchr (sPerlCode, '%') ;	
		}
	    else
		{
		sPerlCode = p ;
		p = NULL ; 
		}
	    }
	if (valid)
	    StringAdd (ppCode, sPerlCode,  0) ; 
	}
    return valid ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileCleanupSpaces                                             */
/*                                                                          */
/* remove any following spaces                                              */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int embperl_CompileCleanupSpaces  (/*in*/  tReq *	r,
					  /*in*/  tDomTree *	pDomTree,
					  /*in*/  tNodeData *	pNode,
					  /*i/o*/ tEmbperlCmd *	pCmd)


    {
    if ((pCmd -> bRemoveNode & 6) && (r -> bOptions & optKeepSpaces) == 0)
	{
	tNodeData *  pNextNode   = Node_selfFirstChild (pDomTree, pNode) ;
	if ((pCmd -> bRemoveNode & 1) || !pCmd -> bCompileChilds || pNextNode == NULL || (pNextNode -> nType != ntypText && pNextNode -> nType != ntypCDATA))
	    pNextNode    = Node_selfNextSibling (pDomTree, pNode) ;
	if (pNextNode)
	    {
	    const char * sText        = Node_selfNodeName (pNextNode) ;
	    const char * p            = sText ;

	    while (*p && isspace (*p))
		p++;
	    if (p > sText && (pCmd -> bRemoveNode & 4))
		p-- ;

	    if (p > sText)
		{ /* remove spaces */
		if (*p)
		    Node_replaceChildWithCDATA(pDomTree, -1, pNextNode -> xNdx, p, strlen (p), -1, 0) ;
		else
		    Node_selfRemoveChild(pDomTree, -1, pNextNode) ;
		}

	    }
	}
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileNode                                                      */
/*                                                                          */
/* Compile one node and his childs                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int embperl_CompileNode (/*in*/  tReq *         r,
			 /*in*/  tDomTree *	pDomTree,
			 /*in*/  tNode		xNode,
			 /*i/o*/ int *		bCheckpointPending)


    {
    int rc ;
    tNode           xChildNode  ;
    tStringIndex    nNdx  ;
    tEmbperlCmd *   pCmd  ;
    tNodeData *     pNode = Node_self (pDomTree, xNode) ;
    tAttrData *     pAttr ;
    int             nAttr = 0 ;
    char *          pCode = NULL ; 
    char *          sStackValue = NULL ;
    int		    nStartCodeOffset = 0 ;               
    int		    nCheckpointCodeOffset = 0 ;               
    char *          sSourcefile ;
    int             nSourcefile ;

    Ndx2StringLen (pDomTree -> xFilename, sSourcefile, nSourcefile) ;

    if (pCurrReq -> bDebug & dbgParse)
	lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d -------> parent=%d node=%d type=%d text=%s\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber, Node_parentNode (pDomTree, pNode -> xNdx), pNode -> xNdx, pNode -> nType, Node_selfNodeName(pNode)) 

; 
    
    //if (*bCheckpointPending && (pNode -> nType == ntypText || pNode -> nType == ntypCDATA) && pNode -> bFlags && (pNode -> bFlags & nflgIgnore) == 0)
    if (*bCheckpointPending && pNode -> bFlags && (pNode -> bFlags & nflgIgnore) == 0)
	{
	int l ;
	char buf [80] ;
	
	pNode -> bFlags |= nflgCheckpoint ;
	l = sprintf (buf, " _ep_cp($_ep_DomTree,%d) ;\n", *bCheckpointPending<0?pNode -> xNdx:*bCheckpointPending) ;
	nCheckpointCodeOffset = StringAdd (pProg, buf,  l) ; 
	*bCheckpointPending = 0 ;

	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Checkpoint\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
	
	}

    pCmd = NULL ;
    
    nNdx = Node_selfNodeNameNdx (pNode) ;

    if (nNdx <= nMaxEmbperlCmd)
	{
	pCmd = &pEmbperlCmds[nNdx] ;
	if (pCmd -> nNodeType != pNode -> nType)
	     pCmd = NULL ;
	}
    else
	pCmd = NULL ;
    
    if (pCmd && pCmd -> nSwitchCodeType == 2)
	pProg = &pProgDef ;

	
    if (pCmd == NULL || (pCmd -> bRemoveNode & 8) == 0)
        { /* calculate attributes before tag, but not when tag should be ignored in output stream */
        while (pAttr = Element_selfGetNthAttribut (pDomTree, pNode, nAttr++))
	    {
            if (pAttr -> bFlags & aflgAttrChilds)
                {
                tNodeData * pChild = Node_selfFirstChild (pDomTree, (tNodeData *)pAttr) ;

                while (pChild)
                    {
                    embperl_CompileNode (r, pDomTree, pChild -> xNdx, bCheckpointPending) ;
                    pChild = Node_selfNextSibling (pDomTree, pChild) ;
                    }
                }                

	    }
        }            
    
    if (pCmd)
        {
        int i ;

        for (i = 0; i < pCmd -> numCompileTimePerlCode; i++)
            {
            if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sCompileTimePerlCode[i], &pCode))
                {
                SV * pSV ;
		int   rc ;

		if (pCode)
                    {
	            int l = ArrayGetSize (pCode) ;
		    if (pCurrReq -> bDebug & dbgParse)
			lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d CompileTimeCode:    %*.*s\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber, l, l, pCode) ; 

		    pSV = newSVpvf("package %s ; \n#line %d \"%s\"\n%*.*s",
			    pCurrReq -> Buf.sEvalPackage, pNode ->  nLinenumber, sSourcefile, l,l, pCode) ;
		    if ((rc = EvalDirect (pCurrReq, pSV)) != ok)
			LogError (pCurrReq, rc) ;
		    SvREFCNT_dec(pSV);
		    }
                break ;
                }
            }
        for (i = 0; i < pCmd -> numPerlCode; i++)
            {
            if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sPerlCode[i], &pCode))
                {
                if (pCode)
                    {
	            int l = ArrayGetSize (pCode) ;

                    if (l)
                        {
                        char buf [32] ;

                        if (pNode ->  nLinenumber)
                            {
                            int l2 = sprintf (buf, "#line %d \"", pNode ->  nLinenumber) ;

                            StringAdd (pProg, buf, l2) ;
                            StringAdd (pProg, sSourcefile, nSourcefile) ;
                            StringAdd (pProg, "\"\n", 2) ;
                            }                        
                    
                        if (pCmd -> bPerlCodeRemove)
                            nStartCodeOffset = StringAdd (pProg, " ", 1) ;
                        StringAdd (pProg, pCode, l) ;
	                StringAdd (pProg, "\n",  1) ; 
			if (pCurrReq -> bDebug & dbgParse)
			    lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Code:    %*.*s\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber, l, l, pCode) ; 
                        }
	            }
                break ;
                }
            }

	embperl_CompileCleanupSpaces (r, pDomTree, pNode, pCmd) ;
        
        if (pCmd -> sMayJump)
            if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sMayJump, &sStackValue))
	        {
		*bCheckpointPending = -1 ;
		if (pCurrReq -> bDebug & dbgParse)
		    lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Set Checkpoint pending\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
	        }

	if (pCmd -> bRemoveNode & 1)
	    Node_selfRemoveChild(pDomTree, -1, pNode) ;
        else if (pCmd -> bRemoveNode & 8)
	    pNode -> bFlags |= nflgIgnore ;

	if (nCheckpointCodeOffset && (pNode -> bFlags == 0 || (pNode -> bFlags & nflgIgnore)))
	    {
	    (*pProg)[nCheckpointCodeOffset] = '#' ;
	    if (pCurrReq -> bDebug & dbgParse)
		lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Remove Checkpoint\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
	    nCheckpointCodeOffset = 0 ;
	    }
    
	if (*bCheckpointPending && (pNode -> bFlags & nflgIgnore))
	    {
	    int l ;
	    char buf [80] ;
	    
	    pNode -> bFlags |= nflgCheckpoint ;
	    l = sprintf (buf, " _ep_cp($_ep_DomTree,%d) ;\n", *bCheckpointPending<0?pNode -> xNdx:*bCheckpointPending) ;
	    nCheckpointCodeOffset = StringAdd (pProg, buf,  l) ; 
	    *bCheckpointPending = 0 ;

	    if (pCurrReq -> bDebug & dbgParse)
		lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Checkpoint\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
	    
	    }

	if (pCmd -> sStackName)
	    {
	    if (pCmd -> sMatchStack && pNode -> nType != ntypStartTag && pNode -> nType != ntypDocument && pNode -> nType != ntypDocumentFraq)
		{
		if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sMatchStack, &sStackValue))
		    if ((rc = embperl_CompileMatchStack (pDomTree, pNode, pCmd -> sStackName, sStackValue)) != ok)
			return rc ;
		}
	    if (pCmd -> sPushStack)
		{
		if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sPushStack, &sStackValue))
		    embperl_CompilePushStack (pDomTree, pCmd -> sStackName, sStackValue) ;
		}
	    }
	if (pCmd -> sStackName2 && pCmd -> sPushStack2)
            {
            if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sPushStack2, &sStackValue))
	        {
                embperl_CompilePushStack (pDomTree, pCmd -> sStackName2, sStackValue) ;
		}
            }
        }
    
    

    if (pCmd == NULL || pCmd -> bCompileChilds)
	{
	xChildNode = pNode -> bFlags?Node_firstChild (pDomTree, xNode):0 ;

	while (xChildNode)
	    {
	    if ((rc = embperl_CompileNode (r, pDomTree, xChildNode, bCheckpointPending)) != ok)
		return rc ;

	    xChildNode  = Node_nextSibling (pDomTree, xChildNode) ;
	    }
	}
	    
    if (pCmd)
	{
        if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sPerlCodeEnd, &pCode))
            {
            if (pCode)
                {
                int l = ArrayGetSize (pCode) ;
                if (l)
                    {
		    StringAdd (pProg, pCode, l) ;
		    StringAdd (pProg, "\n",  1) ; 
		    if (pCurrReq -> bDebug & dbgParse)
			lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Codeend: %*.*s\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber, l, l, pCode) ; 
		    }
		}
            }
	else
	    {
	    if (pCmd -> bPerlCodeRemove && nStartCodeOffset)
		{
		(*pProg)[nStartCodeOffset] = '#' ;
		if (pCurrReq -> bDebug & dbgParse)
		    lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Remove Codeblock\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
		}
	    }
        if (pCmd -> sPerlCodeEnd && pCmd -> sMayJump)
            if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sMayJump, &sStackValue))
	        {
		*bCheckpointPending = -1 ;
		if (pCurrReq -> bDebug & dbgParse)
		    lprintf (pCurrReq, "[%d]EPCOMP: #%d L%d Set Checkpoint pending\n", pCurrReq -> nPid, pNode -> xNdx, pNode -> nLinenumber) ; 
	        }
	if (pCmd -> sStackName  && (pNode -> nType == ntypStartTag || pNode -> nType == ntypDocument || pNode -> nType == ntypDocumentFraq))
	    {
	    if (pCmd -> sMatchStack)
		{
		if (embperl_CompileToPerlCode (pDomTree, pNode, pCmd -> sMatchStack, &sStackValue))
		    {
		    if ((rc = embperl_CompileMatchStack (pDomTree, pNode, pCmd -> sStackName, sStackValue)) != ok)
			return rc ;
		    }
		}
	    else if (pCmd -> sPushStack)
		embperl_CompilePopStack (pDomTree, pCmd -> sStackName) ;
	    }

        if (pCmd -> sStackName2 && pCmd -> sPushStack2)
            embperl_CompilePopStack (pDomTree, pCmd -> sStackName2) ;
  
	if (pCmd -> nSwitchCodeType == 1)
	    pProg = &pProgRun ;
	}


    StringFree (&sStackValue) ;
    StringFree (&pCode) ;

    return ok ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* embperl_CompileDocument                                                  */
/*                                                                          */
/* Compile the whole document                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int embperl_CompileDocument (/*i/o*/ register req * r)


    {
    int		rc ;
    SV *	pSV ;
    SV **	ppSV ;
    tNode	xNode ;
    tDomTree *	pDomTree ;
    tDomTree *	pCurrDomTree ;
    clock_t	cl1 = clock () ;
    clock_t	cl2  ;
    clock_t	cl3  ;
    clock_t	cl4  ;
    int         bCheckpointPending = 0 ;
    STRLEN      l ;


    tainted = 0 ;
    /* Already Parsed ? */

    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, "SRCDOM", 6, 1) ;  
    if (ppSV == NULL)
        return rcHashError ;

    if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
        {
        strncpy (r -> errdat1, SvPV(*ppSV, l), sizeof (r -> errdat1) - 1) ; 
        LogError (r, rcEvalErr) ;
        return rcEvalErr ;
        }
    else if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PVIV)
        {
        char * s1 = SvPV(*ppSV, l) ;
	char * s2 = strchr (s1, '\t') ;
	int    l1 ;
	if (s2)
	    {
	    l1 = s2 - s1 ;
	    if (l1 > sizeof (r -> errdat1) - 1)
		l1 = sizeof (r -> errdat1) - 1 ;
	    
	    strncpy (r -> errdat1, s1, l1) ; 
	    r -> errdat1[l1] = '\0' ;
	    strncpy (r -> errdat2, s2+1, sizeof (r -> errdat2) - 1) ; 
	    }
	else
	    strncpy (r -> errdat1, s1, sizeof (r -> errdat1) - 1) ; 

	r -> Buf.pCurrPos = NULL ;
	r -> Buf.nSourceline = 0 ;
        return SvIVX(*ppSV) ;
        }
    else if (*ppSV == NULL || !SvIOKp (*ppSV) )
	{
	if (*ppSV)
	    SvREFCNT_dec (*ppSV) ;

	if ((rc = ParseFile (r)) != ok)
	    {
	    *ppSV = newSVpvf ("%s\t%s", r -> errdat1, r -> errdat2) ;
	    SvUPGRADE (*ppSV, SVt_PVIV) ;
	    SvIVX (*ppSV) = rc ;
	    if (r -> xCurrDomTree)
		{
		DomTree_delete(DomTree_self(r -> xCurrDomTree)) ;
		r -> xCurrDomTree = 0 ;
		}
	    
	    return rc ;
	    }

	cl2 = clock () ;
	
	pDomTree = DomTree_self (r -> xCurrDomTree) ;
	*ppSV = pDomTree -> pDomTreeSV ;
	/* SvREFCNT_inc (*ppSV) ;*/
	/* lprintf (r, "IV ppSV=%s type=%d\n", *ppSV?"ok":"NULL", *ppSV?SvTYPE (*ppSV):-1) ; */    


	StringNew (&pProgRun, r -> Buf.pFile -> nFilesize / 4) ;
	StringNew (&pProgDef, r -> Buf.pFile -> nFilesize / 4) ;
	pProg = &pProgRun ;

        pDomTree -> pSV = (SV *)newHV () ;
	nLabelCnt = 1 ;
	
	if ((rc = embperl_CompileNode (r, pDomTree, r -> xDocument, &bCheckpointPending)) != ok)
	    {
	    *ppSV = newSVpvf ("%s\t%s", r -> errdat1, r -> errdat2) ;
	    SvUPGRADE (*ppSV, SVt_PVIV) ;
	    SvIVX (*ppSV) = rc ;
	    if (r -> xCurrDomTree)
		{
		DomTree_delete(DomTree_self(r -> xCurrDomTree)) ;
		r -> xCurrDomTree = 0 ;
		}
	    StringFree (&pProgRun) ;
	    StringFree (&pProgDef) ;
	    
	    return rc ;
	    }

        SvREFCNT_dec (pDomTree -> pSV) ;
        pDomTree -> pSV = NULL ;

	StringAdd (&pProgRun, "", 1) ;
	StringAdd (&pProgDef, "", 1) ;

	cl3 = clock () ;

	l = ArrayGetSize (pProgDef) ;
	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (r, "[%d]EPCOMP: AfterCompileTimeCode:    %*.*s\n", r -> nPid, l, l, pProgDef) ; 

	pSV = newSVpvf("package %s ; \n%*.*s", r -> Buf.sEvalPackage, l,l, pProgDef) ;
	if ((rc = EvalDirect (r, pSV)) != ok)
	    LogError (r, rc) ;
	SvREFCNT_dec(pSV);
    
	if ((rc = EvalStore (r, pProgRun, 0, &pSV)) != ok)
	    return rc ;
    
	cl4 = clock () ;
#ifdef CLOCKS_PER_SEC
        if (r -> bDebug)
	    {
	    lprintf (r, "[%d]PERF: Compile Start Time:	    %d ms \n", r -> nPid, ((cl1 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	    lprintf (r, "[%d]PERF: Parse End Time:	    %d ms \n", r -> nPid, ((cl2 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	    lprintf (r, "[%d]PERF: To Perl End Time:	    %d ms \n", r -> nPid, ((cl3 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	    lprintf (r, "[%d]PERF: Perl Compile End Time:   %d ms \n", r -> nPid, ((cl4 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	    lprintf (r, "[%d]PERF: Compile Time:	    %d ms \n", r -> nPid, ((cl4 - cl1) * 1000 / CLOCKS_PER_SEC)) ;
	    }
#endif        

	StringFree (&pProgRun) ;
	StringFree (&pProgDef) ;
	}
    else
	{
	pDomTree = DomTree_self (SvIVX (*ppSV)) ;
	r -> xDocument = pDomTree -> xDocument ;
	if ((rc = EvalStore (r, pProgRun, 0, &pSV)) != ok)
	    return rc ;
	}


    if (!r -> bError)
	{
	rc = ok ;
	cl1 = clock () ;
	
	if (!(r -> xCurrDomTree  = DomTree_clone (pDomTree, &pCurrDomTree)))
	    return 1 ;


	av_push (r -> pDomTreeAV, pCurrDomTree -> pDomTreeSV) ;

	r -> xDocument = pCurrDomTree -> xDocument ;

	if (pSV)
	    {
	    SV * args[2] ;

	    nCheckpointCache = 0 ;
	    xCheckpointCache[nCheckpointCache++] = 0 ;
	    xCheckpointCache[nCheckpointCache++] = 0 ;
	    
	    args[0] = r -> pReqSV ;
	    args[1] = pCurrDomTree -> pDomTreeSV ;
	    rc = CallStoredCV (r, pProgRun, (CV *)pSV, 2, args, 0, &pSV) ;
	    if (pSV)
		SvREFCNT_dec (pSV) ;

	    cl2 = clock () ;
    #ifdef CLOCKS_PER_SEC
	    if (r -> bDebug)
	        {
	        lprintf (r, "[%d]PERF: Run Start Time: %d ms \n", r -> nPid, ((cl1 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	        lprintf (r, "[%d]PERF: Run End Time:   %d ms \n", r -> nPid, ((cl2 - r -> startclock) * 1000 / CLOCKS_PER_SEC)) ;
	        lprintf (r, "[%d]PERF: Run Time:       %d ms \n", r -> nPid, ((cl2 - cl1) * 1000 / CLOCKS_PER_SEC)) ;
	        }
    #endif    

	    }

        if (rc != ok && rc != rcEvalErr)
            return rc ;

	}
    
    return ok ;
    }
