/*###################################################################################
#
#   Embperl - Copyright (c) 1997-2000 Gerald Richter / ECOS
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.
#
#   THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#   WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
###################################################################################*/


#include "ep.h"
#include "epmacro.h"

HV * pStringTableHash ;	    /* Hash to translate strings to index number */
HE * * pStringTableArray  ;   /* Array with pointers to strings */
tStringIndex  * pFreeStringsNdx  ;   /* List of freed string indexes */


tDomTree * pDomTrees ;
tIndex *   pFreeDomTrees ;

int nInitialNodePadSize = 196 ;

int nMemUsage = 0 ;
int numNodes  = 0 ;
int numAttr   = 0 ;
int numStr    = 0 ;
int numPads   = 0 ;

tIndex xNoName  = 0 ;
tIndex xDomTreeAttr = 0 ;

int nCheckpointCache = 0 ;
int nCheckpointCacheMask = 0x1ff ;
tIndex xCheckpointCache[512] ;


#define dom_malloc(s)  (nMemUsage += s, malloc(s))
#define dom_free(s)    (free(s))


/* forward */
static int DomTree_free (SV * pSV, MAGIC * mg) ;


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* mydie                                                                    */
/*                                                                          */
/* Fatal Error                                                              */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int mydie (char *  msg)
    {
    strncpy (pCurrReq -> errdat1, msg, sizeof (pCurrReq -> errdat1)) ;
    LogError (pCurrReq, 9999) ;
    puts (msg) ;
    exit (1) ;
    }



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArrayNew                                                                 */
/*                                                                          */
/* Create a new dynamic array                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int ArrayNew (/*in*/ const tArray * pArray,
	      /*in*/ int	nAdd,
	      /*in*/ int	nElementSize)


    {
    struct tArrayCtrl * pNew ;
    
    if ((pNew = dom_malloc (nAdd * nElementSize + sizeof (struct tArrayCtrl))) == NULL)
	return 0 ;
    
    memset (pNew, 0, nAdd * nElementSize + sizeof (struct tArrayCtrl)) ; 
    *(void * *)pArray = (struct tArray *)(pNew + 1) ;
    pNew -> nMax = nAdd ;
    pNew -> nAdd = nAdd ;
    pNew -> nFill = 0  ;
    pNew -> nElementSize = nElementSize  ;

    return ok ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArrayFree                                                                */
/*                                                                          */
/* Create a new dynamic array                                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int ArrayFree (/*in*/ const tArray * pArray)


    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pArray)) - 1 ;

    if (pCtrl)
        dom_free (pCtrl) ;

    (*(void * *)pArray) = NULL ;

    return ok ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArrayClone                                                               */
/*                                                                          */
/* Create a new dynamic array as exact copy of old one                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int ArrayClone (/*in*/  const tArray * pOrgArray,
	        /*out*/ const tArray * pNewArray)


    {
    struct tArrayCtrl * pNew ;
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pOrgArray)) - 1 ;
    int    size = pCtrl -> nFill * pCtrl -> nElementSize + sizeof (struct tArrayCtrl) ;
    
    if ((pNew = dom_malloc (size)) == NULL)
	return 0 ;
    
    memcpy (pNew, pCtrl, size) ; 
    *(void * *)pNewArray = (struct tArray *)(pNew + 1) ;
    pNew -> nMax = pCtrl -> nFill ;

    return ok ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArrayAdd                                                                 */
/*                                                                          */
/* Make space for numElements in Array and return index of first one        */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int ArrayAdd (/*in*/ const tArray * pArray,
	      /*in*/ int	numElements)

    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pArray)) - 1 ;
    int	         nNdx ;


    if (pCtrl -> nFill + numElements > pCtrl -> nMax)
	{
	struct tArrayCtrl * pNew ;
	int		    nNewMax = pCtrl -> nFill + numElements + pCtrl -> nAdd ;
	
	if ((pNew = realloc (pCtrl, nNewMax * pCtrl -> nElementSize + sizeof (struct tArrayCtrl))) == NULL)
	    return 0 ;
	
	*(void * *)pArray = (struct tArray *)(pNew + 1) ;
	pNew -> nMax = nNewMax ;
	pCtrl = pNew ;
	}

    nNdx = pCtrl -> nFill ;
    pCtrl -> nFill += numElements ;
    return nNdx ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArraySub								    */
/*                                                                          */
/* Remove numElemenets from the end of the array. Does not reallocate memory*/
/* Returns -1 if less then numElemenets are avialable			    */
/*									    */
/* ------------------------------------------------------------------------ */


int ArraySub (/*in*/ const tArray * pArray,
	      /*in*/ int	numElements)

    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pArray)) - 1 ;
    int	         nNdx ;


    if (pCtrl -> nFill < numElements)
	return -1 ;
    else
	pCtrl -> nFill -= numElements ;

    return pCtrl -> nFill ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArraySet                                                                 */
/*                                                                          */
/* Make space that at least numElements in the Array                        */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int ArraySet (/*in*/ const tArray * pArray,
	      /*in*/ int	numElements)

    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pArray)) - 1 ;
    int	         nNdx ;
    char *       p ;

    if (numElements > pCtrl -> nMax)
	{
	struct tArrayCtrl * pNew ;
	int		    nNewMax = pCtrl -> nFill + pCtrl -> nAdd ;

	if (nNewMax < numElements)
	    nNewMax = numElements + pCtrl -> nAdd ;
	
	if ((pNew = realloc (pCtrl, nNewMax * pCtrl -> nElementSize + sizeof (struct tArrayCtrl))) == NULL)
	    return 0 ;
	
	p = (char *)(pNew + 1) ;
	*(void * *)pArray = (struct tArray *)p ;
	memset (p + pNew -> nMax * pNew -> nElementSize, 0, (nNewMax - pNew -> nMax) * pNew -> nElementSize) ; 
	pNew -> nMax = nNewMax ;
	pCtrl = pNew ;
	}

    return numElements ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArraySetSize                                                             */
/*                                                                          */
/* Make space for exact numElements in the Array			    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int ArraySetSize (/*in*/ const tArray * pArray,
	          /*in*/ int	numElements)

    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(*(void * *)pArray)) - 1 ;
    int	         nNdx ;
    char *       p ;

    
    if (numElements > pCtrl -> nMax)
	ArraySet (pArray, numElements) ;
    
    pCtrl -> nFill = numElements ;

    return numElements ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ArrayGetSize                                                             */
/*                                                                          */
/* Get size of Array							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int ArrayGetSize (/*in*/ const tArray * pArray)

    {
    struct tArrayCtrl * pCtrl = ((struct tArrayCtrl *)(pArray)) - 1 ;
    
    return pCtrl -> nFill  ;
    }



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* StringNew                                                                */
/*                                                                          */
/* create a new string                                                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */


void StringNew (/*in*/ char * * pArray,
	      /*in*/ int	nAdd)

    {	      
    if ((*(void * *)pArray) == NULL)
	ArrayNew (pArray, nAdd, sizeof (char)) ;
    else
	ArraySetSize (pArray, 0);
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* StringFree                                                               */
/*                                                                          */
/* Free String memory                                                       */
/*                                                                          */
/* ------------------------------------------------------------------------ */


void StringFree (/*in*/ char * * pArray)
	         

    {	      
    if ((*(void * *)pArray) != NULL)
	ArrayFree (pArray) ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* StringAdd                                                                */
/*                                                                          */
/* append to string                                                         */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int StringAdd (/*in*/ char * *	   pArray,
	       /*in*/ const char * sAdd,
	       /*in*/ int 	   nLen)

    {	      
    int nIndex ;
    
    if (nLen == 0)
	nLen = strlen (sAdd) ;

    nIndex = ArrayAdd (pArray, nLen) ;

    memcpy ((*pArray)+nIndex, sAdd, nLen) ;
    
    return nIndex ;
    }




/* ------------------------------------------------------------------------ */
/*                                                                          */
/* String2Ndx                                                               */
/*                                                                          */
/* Convert String to an unique index                                        */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tStringIndex String2NdxInc (/*in*/ const char *	    sText,
			    /*in*/ int		    nLen,
			    /*in*/ int              bInc) 

    {
    SV * *  ppSV ;
    SV *    pSVKey ;
    SV *    pSVNdx ;
    HE *    pHEKey ;
    int	    nNdx ;
    

    if (sText == NULL)
	return 0 ;

    if ((ppSV = hv_fetch (pStringTableHash, (char *)sText, nLen, 0)) != NULL)
	{

	/*lprintf (pCurrReq, "String2Ndx type=%d  iok=%d flg=%x\n", *ppSV?SvTYPE(*ppSV):-1, SvIOK (*ppSV), SvFLAGS(*ppSV)) ;*/

	if (*ppSV != NULL && SvIOKp (*ppSV)) /* use SvIOKp to avoid problems with tainting */
	    {
	    if (bInc)
		SvREFCNT_inc (*ppSV) ;
	    return SvIVX (*ppSV) ;
	    }
	}


    /* new string */
	
    nNdx = ArraySub (&pFreeStringsNdx, 1) ;
    if (nNdx != (tIndex)(-1))
	nNdx = pFreeStringsNdx[nNdx] ;
    else
	nNdx = ArrayAdd (&pStringTableArray, 1) ;
    
    pSVNdx = newSViv (nNdx) ;
    SvTAINTED_off (pSVNdx) ;

    if (bInc)
	SvREFCNT_inc (pSVNdx) ;  
    pSVKey = newSVpv (nLen?(char *)sText:"", nLen) ;
    pHEKey = hv_store_ent (pStringTableHash, pSVKey, pSVNdx, 0) ;
    SvREFCNT_dec (pSVKey) ;

    pStringTableArray[nNdx] = pHEKey ;

    numStr++ ;

    return nNdx ;    
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NdxStringFree                                                            */
/*                                                                          */
/* Free one reference to a string                                           */
/*                                                                          */
/* ------------------------------------------------------------------------ */


void NdxStringFree (/*in*/ tStringIndex		    nNdx)

    {
    HE *    pHE = pStringTableArray[nNdx] ;
    SV *    pSVNdx = HeVAL (pHE) ;
    
    SvREFCNT_dec (pSVNdx) ;

    /*
    if (nNdx == 3)
	lprintf (pCurrReq, "free string %s (#%d) refcnt=%d\n", Ndx2String (nNdx), nNdx, SvREFCNT(pSVNdx)) ;
    */
    
    if (SvREFCNT(pSVNdx) == 1)
	{
	int n ;
	
	/* lprintf (pCurrReq, "delete string %s (#%d)\n", Ndx2String (nNdx), nNdx) ; */
	hv_delete (pStringTableHash, HeKEY (pHE), HeKLEN(pHE), 0) ;
	pStringTableArray[nNdx] = NULL ;
	n = ArrayAdd (&pFreeStringsNdx, 1) ;
	pFreeStringsNdx[n] = nNdx ;

	numStr-- ;
	}
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomInit                                                                  */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int DomInit (void)

    {
    SV *    pSVKey ;
    SV *    pSVNdx ;
    HE *    pHEKey ;
    
    pStringTableHash = newHV () ;

    ArrayNew (&pStringTableArray, 256, sizeof (HE *)) ; 
    ArrayNew (&pFreeStringsNdx, 256, sizeof (tStringIndex *)) ; 

    ArrayAdd (&pStringTableArray, 2) ;
    /* NULL */
    pSVNdx = newSViv (0) ;
    SvREFCNT_inc (pSVNdx) ;  
    pSVKey = newSVpv ("", 0) ;
    pHEKey = hv_store_ent (pStringTableHash, pSVKey, pSVNdx, 0) ;
    pStringTableArray[0] = pHEKey ;

    /* "" */
    pSVNdx = newSViv (1) ;
    SvREFCNT_inc (pSVNdx) ;  
    pSVKey = newSVpv ("", 0) ;
    pHEKey = hv_store_ent (pStringTableHash, pSVKey, pSVNdx, 0) ;
    pStringTableArray[1] = pHEKey ;

    numStr+=2 ;

    xNoName  = String2Ndx ("<noname>", 8) ;
    xDomTreeAttr = String2Ndx ("<domtree>", 9) ;

    ArrayNew (&pDomTrees, 16, sizeof (tDomTree)) ; 
    ArrayAdd (&pDomTrees, 1) ;
    memset (&pDomTrees[0], 0, sizeof (tDomTree)) ;
    ArrayNew (&pFreeDomTrees, 16, sizeof (tIndex)) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_alloc							    */
/*                                                                          */
/* Alloc storage for a new DomTree and associate it with a SV. deleteing    */
/* the SV will delete the DomTree  also 				    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

MGVTBL DomTree_mvtTab = { NULL, NULL, NULL, NULL, DomTree_free } ;


tDomTree * DomTree_alloc (void)

    {
    tDomTree * pDomTree ;
    tIndex     n ;
    SV *       pSV ;
    struct magic * pMagic ;


    n = ArraySub (&pFreeDomTrees, 1) ;
    if (n != (tIndex)(-1))
	n = pFreeDomTrees[n] ;
    else
	n = ArrayAdd (&pDomTrees, 1) ;
    pDomTree = DomTree_self (n) ;

    memset (pDomTree, 0, sizeof (*pDomTree)) ;
    
    pSV = newSViv (n) ;
    sv_magic (pSV, pSV, 0, NULL, n) ;
    pMagic = mg_find (pSV, 0) ;

    if (pMagic)
	pMagic -> mg_virtual = &DomTree_mvtTab ;
    else
        {
        LogError (pCurrReq, rcMagicError) ;
        }

    pDomTree -> pDomTreeSV = pSV ;
    pDomTree -> xNdx = n ;


    return pDomTree ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_new                                                              */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int DomTree_new (tDomTree * * pNewLookup)

    {
    tDomTree * pDomTree ;
    pDomTree = DomTree_alloc () ;

    ArrayNew (&pDomTree -> pLookup, 128, sizeof (struct tNodeData *)) ; 
    ArrayAdd (&pDomTree -> pLookup, 1) ;

    ArrayNew (&pDomTree -> pOrder, 128, sizeof (tDomTreeOrder)) ; 

    *pNewLookup = pDomTree  ;

    return pDomTree -> xNdx ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_dodelete							      */
/*                                                                          */
/* Frees all memory allocated by this DomTree				    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

static int DomTree_dodelete (tDomTree * pDomTree)

    {
    tNodePad * * pLookup = (tNodePad * *)pDomTree -> pLookup ;
    int        numLookup  ;
    tIndex     xDomTree  = pDomTree -> xNdx ;
    tIndex     xNdx ;

    if (pCurrReq -> bDebug & dbgParse)
	lprintf (pCurrReq, "[%d]Delete: DomTree = %d SVs=%d\n", pCurrReq -> nPid, pDomTree -> xNdx, sv_count) ; 

    if (!xDomTree)
	{
	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (pCurrReq, "[%d]Delete: Already deleted DomTree = %d SVs=%d\n", pCurrReq -> nPid, pDomTree -> xNdx, sv_count) ; 
	return ok ;
	}
    
    numLookup = ArrayGetSize (pLookup) ;
    pLookup += numLookup - 1 ;
    while (numLookup-- > 0)
	{
	tNodePad * pPad = *pLookup ;
	
	if (pPad && pPad -> nType == (tNodeType)ntypPad && xDomTree == pPad -> xDomTree)
	    {
	    tNodeData * pNode = NodePad_selfFirstChild (pDomTree,pPad) ;
	    int         nOffset ;

	    /* lprintf (pCurrReq, "delete typ=%d  Pad  xNdx=%d\n", pPad -> nType, pPad -> xNdx) ; */

	    while (pNode)
		{
		int n = pNode -> numAttr ;
		tAttrData * pAttr = Node_selfFirstAttr(pNode) ;

		while (n--)
		    {
		    pDomTree -> pLookup[pAttr -> xNdx] = NULL ;
		    if (pAttr -> bFlags)
			{
			if (pAttr -> xName)
			    NdxStringFree (pAttr -> xName) ;
			if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
			    NdxStringFree (pAttr -> xValue) ;
			}
		    pAttr++ ;
		    }
		/* lprintf (pCurrReq, "delete typ=%d  Node xNdx=%d\n", pNode -> nType, pNode -> xNdx) ; */
		pDomTree -> pLookup[pNode -> xNdx] = NULL ;
		if (pNode -> nText)
		    NdxStringFree (pNode -> nText) ;
		
		nOffset = sizeof (struct tNodeData) + sizeof (struct tAttrData) * pNode -> numAttr ;
		if (pPad -> nFill <= pNode -> nPadOffset + nOffset) 
		    break ;
		pNode = (struct tNodeData *)(((tUInt8 *)pNode) + nOffset) ;
		}
	    dom_free (pPad) ;
	    }

	pLookup-- ;
	}

    ArrayFree (&pDomTree -> pLookup) ;
    ArrayFree (&pDomTree -> pOrder) ;
    
    if (pDomTree -> pSV)
	SvREFCNT_dec (pDomTree -> pSV) ;


    xNdx = ArrayAdd (&pFreeDomTrees, 1) ;
    pDomTree -> xNdx = 0 ;
    pFreeDomTrees[xNdx] = xDomTree ;

    return ok ;    
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_free 							    */
/*                                                                          */
/* Frees all memory allocated by this DomTree				    */
/* Do not call directly. Is called by Perl when RefCnt goes to zero	    */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static int DomTree_free (SV * pSV, MAGIC * mg)

    {
    return DomTree_dodelete (DomTree_self (mg -> mg_len)) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_delete							    */
/*                                                                          */
/* Frees all memory allocated by this DomTree				    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int DomTree_delete (tDomTree * pDomTree)

    {
    SvREFCNT_dec (pDomTree -> pDomTreeSV) ;
    return ok ;
    }

    
/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_clone                                                            */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int DomTree_clone (/*in*/ tDomTree *	pOrgDomTree,
		   /*out*/tDomTree * *  pNewDomTree)

    {
    tDomTree * pDomTree ;
    tNodeData * pDocument ;
    tIndex      xOrgDomTree = pOrgDomTree -> xNdx ; 
    
    	
    pDomTree = DomTree_alloc () ;
    pOrgDomTree = DomTree_self (xOrgDomTree) ; /* relookup in case it has moved */
    
    pDomTree -> xDocument = pOrgDomTree -> xDocument ;
    pDomTree -> xFilename = pOrgDomTree -> xFilename ;

    ArrayClone (pOrgDomTree, &pDomTree -> pLookup) ; 
    ArrayNew (&pDomTree -> pOrder, 128, sizeof (tDomTreeOrder)) ; 

    if (pDomTree -> pSV = pOrgDomTree -> pSV)
        SvREFCNT_inc (pDomTree -> pSV) ;

    pDocument = Node_self (pDomTree, pDomTree -> xDocument) ;
    
    if (pDocument -> nType == ntypDocumentFraq)
	{
	tAttrData * pAttr; 
	pDocument = Node_selfCloneNode (pDomTree, pDocument, 1) ;
	pAttr = Element_selfSetAttribut (pDomTree, pDocument, NULL, xDomTreeAttr, NULL, pDomTree -> xNdx, 0) ;
	pAttr -> bFlags = aflgOK ; /* reset string value flag */
	pDomTree -> xDocument = pDocument -> xNdx ;
	}
    

    *pNewDomTree = pDomTree  ;

    return pDomTree -> xNdx ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_checkpoint                                                       */
/*                                                                          */
/* Add a new checkpoint to sync the programm execution with the DOM tree    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

void DomTree_checkpoint (tIndex xDomTree, tNode xChild)


    {
    tDomTree * pDomTree = DomTree_self (xDomTree) ;
    int n = ArrayAdd (&pDomTree -> pOrder, 1) ;
    (pDomTree -> pOrder)[n].xFromNode = xChild ;
    (pDomTree -> pOrder)[n].xToNode   = 0 ;
    if (pCurrReq -> bDebug & dbgParse)
	lprintf (pCurrReq, "[%d]Checkpoint: Node=%d OrderIndex %d SVs=%d\n", pCurrReq -> nPid, xChild, n, sv_count) ; 
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_selfCheckpoint                                                   */
/*                                                                          */
/* Add a new checkpoint to sync the programm execution with the DOM tree    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

void DomTree_selfCheckpoint (tDomTree * pDomTree, tNode xFromNode, tNode xToNode)


    {
    int n = ArrayAdd (&pDomTree -> pOrder, 1) ;
    (pDomTree -> pOrder)[n].xFromNode = xFromNode ;
    (pDomTree -> pOrder)[n].xToNode   = xToNode ;
    if (pCurrReq -> bDebug & dbgParse)
	lprintf (pCurrReq, "[%d]Checkpoint: Node=%d -> %d OrderIndex %d SVs=%d\n", pCurrReq -> nPid, xFromNode, xToNode, n, sv_count) ; 
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* DomTree_discardAfterCheckpoint                                           */
/*                                                                          */
/* Search for the next checkpoint after xNode and discard all checkpoints   */
/* set after and including that one					    */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int DomTree_selfDiscardAfterCheckpoint (/*in*/ tDomTree *	    pDomTree, 
				         /*in*/ tNodeData *     pNode,
				         /*in*/ tNodeData *     pArgNode)


    {
    tNode	xNode ;
    tNodeData * pChild ;    
    tDomTreeOrder *  pOrder ;
    int         n ;
    int         o ;


    if (!pArgNode)
	{
    	pArgNode = pNode ;
	pNode = Node_selfFirstChild (pDomTree, pNode) ;
	if (!pNode)
	    pNode = Node_selfNextSibling (pDomTree, pNode) ;
	}
    
    while (pNode && (pNode -> bFlags & nflgCheckpoint) == 0) 
	{
	pChild = Node_selfFirstChild (pDomTree, pNode) ;
	if (pChild)
	    if (DomTree_selfDiscardAfterCheckpoint (pDomTree, pChild, pArgNode))
		return 1 ;
				    
	pNode = Node_selfNextSibling (pDomTree, pNode) ;
	}

    if (pNode == NULL)
	return 0 ;

    xNode = pNode -> xNdx ;

    pOrder = pDomTree -> pOrder ;
    o = n = ArrayGetSize (pOrder) ;
    while (--n >= 0)
	{
	if (pOrder[n].xFromNode == xNode)
	    break ;
	}
    ArraySetSize (&pDomTree -> pOrder, n) ;

    if (pCurrReq -> bDebug & dbgParse)
	lprintf (pCurrReq, "[%d]Checkpoint Discard: Node=%d Checkpoint=%d  OrderIndex %d -> %d\n", pCurrReq -> nPid, pArgNode -> xNdx, xNode, o, n) ; 
    
    return 1 ;
    }



int DomTree_discardAfterCheckpoint (/*in*/ tIndex  xDomTree, 
				     /*in*/ tNode   xNode)


    {
    tDomTree * pDomTree = DomTree_self (xDomTree) ;
    tNodeData * pNode   = Node_self (pDomTree, xNode) ;
    
    return DomTree_selfDiscardAfterCheckpoint (pDomTree, pNode, NULL) ;
    }

/* ------------------------------------------------------------------------ 

interface Node {
  // NodeType
  const unsigned short      ELEMENT_NODE                   = 1;
  const unsigned short      ATTRIBUTE_NODE                 = 2;
  const unsigned short      TEXT_NODE                      = 3;
  const unsigned short      CDATA_SECTION_NODE             = 4;
  const unsigned short      ENTITY_REFERENCE_NODE          = 5;
  const unsigned short      ENTITY_NODE                    = 6;
  const unsigned short      PROCESSING_INSTRUCTION_NODE    = 7;
  const unsigned short      COMMENT_NODE                   = 8;
  const unsigned short      DOCUMENT_NODE                  = 9;
  const unsigned short      DOCUMENT_TYPE_NODE             = 10;
  const unsigned short      DOCUMENT_FRAGMENT_NODE         = 11;
  const unsigned short      NOTATION_NODE                  = 12;

  readonly attribute DOMString        nodeName;
           attribute DOMString        nodeValue;
                                        // raises(DOMException) on setting
                                        // raises(DOMException) on retrieval

  readonly attribute unsigned short   nodeType;
  readonly attribute Node             parentNode;
  readonly attribute NodeList         childNodes;
  readonly attribute Node             firstChild;
  readonly attribute Node             lastChild;
  readonly attribute Node             previousSibling;
  readonly attribute Node             nextSibling;
  readonly attribute NamedNodeMap     attributes;
  // Modified in DOM Level 2:
  readonly attribute Document         ownerDocument;
  Node               insertBefore(in Node newChild, 
                                  in Node refChild)
                                        raises(DOMException);
  Node               replaceChild(in Node newChild, 
                                  in Node oldChild)
                                        raises(DOMException);
  Node               removeChild(in Node oldChild)
                                        raises(DOMException);
  Node               appendChild(in Node newChild)
                                        raises(DOMException);
  boolean            hasChildNodes();
  Node               cloneNode(in boolean deep);
  // Introduced in DOM Level 2:
  void               normalize();
  // Introduced in DOM Level 2:
  boolean            supports(in DOMString feature, 
                              in DOMString version);
  // Introduced in DOM Level 2:
  readonly attribute DOMString        namespaceURI;
  // Introduced in DOM Level 2:
           attribute DOMString        prefix;
                                        // raises(DOMException) on setting

  // Introduced in DOM Level 2:
  readonly attribute DOMString        localName;
};



*/

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NewPad                                                                   */
/*                                                                          */
/* Create a new pad for storing nodes                                       */
/*                                                                          */
/* ------------------------------------------------------------------------ */


static struct tNodePad * NewPad (/*in*/ tDomTree *  pDomTree,
				 /*in*/ tIndex	    xParent,
				 /*in*/ int         nSize)

    {
    tIndex xNdx = ArrayAdd (&pDomTree -> pLookup, 1) ;
    struct tNodePad * pChilds ;

    if (nSize == 0)
	nSize = nInitialNodePadSize ;
    else
	nSize = (nSize + sizeof (struct tNodePad) / nInitialNodePadSize + 1) * nInitialNodePadSize ;
    
    if ((pDomTree -> pLookup[xNdx] = pChilds = dom_malloc (nSize)) == NULL)
	return NULL ;
    
    pChilds -> nType = ntypPad ;
    pChilds -> nFill = sizeof (struct tNodePad) ;
    pChilds -> nMax  = nInitialNodePadSize  ;
    pChilds -> numChilds = 0 ;
    pChilds -> xParent = xParent ;
    pChilds -> xNdx    = xNdx ;
    pChilds -> xFirst  = xNdx ;
    pChilds -> xNext   = 0 ;
    pChilds -> xPrev   = 0 ;
    pChilds -> xLast   = 0 ;
    pChilds -> xDomTree = pDomTree -> xNdx ;


    numPads++ ;
    return pChilds ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NodePad_appendChild                                                      */
/*                                                                          */
/* Append a child node to a pad                                             */
/*                                                                          */
/* ------------------------------------------------------------------------ */

tNodeData *  NodePad_appendChild   (/*in*/ tDomTree *   pDomTree,
			            /*in*/ tNodePad * * ppPad,
			            /*in*/ tNodeType	nType,
			            /*in*/ tIndex       xText,
                                    /*in*/ int          nLinenumber)

    {
    tNodePad *  pPad = *ppPad ;
    tNodeData * pNew  ;
    tNode xNdx  ;

    if (pPad -> xLast && pPad -> xFirst == pPad -> xNdx)
	pPad = NodePad_self (pDomTree, pPad -> xLast) ;
	    	    
    if (pPad -> nFill + sizeof (struct tNodeData) > pPad -> nMax)
	{ /* pad full -> make room */
	struct tNodePad *  pNext = NewPad (pDomTree, pPad -> xParent, 0) ;
	struct tNodePad * pFirst = NodePad_self (pDomTree, pPad -> xFirst) ;

	pNext   -> xPrev  = pPad    -> xNdx ;
	pPad    -> xNext  = pNext   -> xNdx ;
	pNext   -> xFirst = pFirst  -> xFirst ;
	if (pFirst -> xLast == pPad -> xNdx || pFirst -> xLast == 0)
	    pFirst -> xLast = pNext -> xNdx ;

	*ppPad = pPad = pNext ;
	}

    pNew = (struct tNodeData *)(((tUInt8 *)pPad) + pPad -> nFill) ;
    xNdx = ArrayAdd (&pDomTree -> pLookup, 1) ;

    pDomTree -> pLookup[xNdx] = pNew ;

    pNew -> nText   = xText ;
    pNew -> nType   = nType ;
    pNew -> numAttr = 0 ;
    pNew -> xNdx    = xNdx ;
    pNew -> xChilds = 0 ;
    pNew -> bFlags  = nflgOK ;
    pNew -> nPadOffset = pPad -> nFill ;
    pNew -> nLinenumber = nLinenumber ;

    pPad -> numChilds++ ;
    pPad -> nFill += sizeof (struct tNodeData) ;
    numNodes++ ;

    return pNew ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NodePad_splitPad                                                         */
/*                                                                          */
/* Split a pad into two                                                     */
/*                                                                          */
/* ------------------------------------------------------------------------ */

tNodePad *  NodePad_splitPad       (/*in*/ tDomTree *   pDomTree,
			            /*in*/ tNodePad *   pPad,
			            /*in*/ tNodeData *	pNewFirstNode,
				    /*in*/ int          nNewNodeSize)

    {
    struct tNodePad *  pFirst;
    struct tNodePad *  pNext      = NewPad (pDomTree, pPad -> xParent, nNewNodeSize) ;
    struct tNodeData * pNewNode   = NodePad_selfFirstChild (pDomTree, pNext) ;
    int                nCopySize ;
    int                nOffset ;
    void * *    pLookup  ;


    if (pPad -> numChilds == 1)
	{
	pNext -> xPrev  = pPad -> xPrev ;
	pNext -> xNext  = pPad -> xNext ;
	pNext -> xFirst = pPad -> xFirst ;
	pNext -> xLast  = pPad -> xLast ;
	pNext -> xNdx   = pPad -> xNdx ;
	ArraySetSize (&pDomTree -> pLookup, ArrayGetSize (pDomTree -> pLookup) - 1) ;
	pDomTree -> pLookup[pNext -> xNdx] = pNext ;
	}
    else
	{
	pNext -> xPrev  = pPad -> xNdx ;
	pNext -> xNext  = pPad -> xNext ;
	pNext -> xFirst = pPad -> xFirst ;
    
	pFirst = NodePad_self (pDomTree, pPad -> xFirst) ;
	if (pFirst -> xLast == pPad -> xNdx || pFirst -> xLast == 0)
	    pFirst -> xLast = pNext -> xNdx ;

	pPad  -> xNext = pNext -> xNdx ;
	}

    nCopySize = pPad  -> nFill - pNewFirstNode -> nPadOffset ;
    memcpy (pNewNode, pNewFirstNode, nCopySize) ;
    pPad  -> nFill -= nCopySize ;
    pNext -> nFill += nCopySize ;

    pLookup  = pDomTree -> pLookup ;
    while (pNewNode)
	{
	int n = pNewNode -> numAttr ;
	tAttrData * pAttr = Node_selfFirstAttr(pNewNode) ;

	while (n--)
	    {
	    pLookup[pAttr -> xNdx] = pAttr ;
	    pAttr++ ;
	    }

	pNewNode -> nPadOffset = ((tUInt8 *)pNewNode) - ((tUInt8 *)pNext) ;
	pLookup[pNewNode -> xNdx] = pNewNode ;
	pPad -> numChilds-- ;
	pNext -> numChilds++ ;
	nOffset = sizeof (struct tNodeData) + sizeof (struct tAttrData) * pNewNode -> numAttr ;
	if (pNext -> nFill <= pNewNode -> nPadOffset + nOffset) 
	    break ;
	pNewNode = (struct tNodeData *)(((tUInt8 *)pNewNode) + nOffset) ;
	}

    if (pPad -> numChilds == 0)
	{
	free (pPad) ;
	}
    
    return pNext ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NodePad_clonePad                                                         */
/*                                                                          */
/* clone a pad                                                              */
/*                                                                          */
/* ------------------------------------------------------------------------ */

tNodePad *  NodePad_clonePad       (/*in*/ tDomTree *   pDomTree,
			            /*in*/ tNodePad *   pPad)

    {
    tNodePad *  pNewPad ;
    tIndex      xNdx  = pPad -> xNdx ;
    tNodeData * pNewNode ;
    int         nOffset ;
    void * *    pLookup  = pDomTree -> pLookup ;

    if ((pLookup[xNdx] = pNewPad = dom_malloc (pPad -> nMax)) == NULL)
	return NULL ;

    memcpy (pNewPad, pPad, pPad -> nMax) ;

    pNewPad -> xDomTree = pDomTree -> xNdx ;

    pNewNode   = NodePad_selfFirstChild (pDomTree, pNewPad) ;

    while (pNewNode)
	{
	int n = pNewNode -> numAttr ;
	tAttrData * pAttr = Node_selfFirstAttr(pNewNode) ;

	if (pNewNode -> nText)
	    NdxStringRefcntInc (pNewNode -> nText) ;
	while (n--)
	    {
	    pLookup[pAttr -> xNdx] = pAttr ;
	    if (pAttr -> xName)
		NdxStringRefcntInc (pAttr -> xName) ;
	    if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
		NdxStringRefcntInc (pAttr -> xValue) ;
	    pAttr++ ;
	    }

	pLookup[pNewNode -> xNdx] = pNewNode ;
	nOffset = sizeof (struct tNodeData) + sizeof (struct tAttrData) * pNewNode -> numAttr ;
	if (pNewPad -> nFill <= pNewNode -> nPadOffset + nOffset) 
	    break ;
	pNewNode = (struct tNodeData *)(((tUInt8 *)pNewNode) + nOffset) ;
	}

    return pNewPad ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_condClonePad                                                        */
/*                                                                          */
/* clone the pad pointed to by node if it lives in another DomTree          */
/*                                                                          */
/* ------------------------------------------------------------------------ */

tNodePad *  Node_condClonePad       (/*in*/ tDomTree *   pDomTree,
	     		             /*in*/ tNode        xNode)

    {
    tNodeData *	pNode = Node_self (pDomTree, xNode) ;
    tNodePad *  pPad  = NULL ;

    if (pNode)
	{
	if (pNode -> nType == ntypAttr)
	    pPad = Node_selfPad (Attr_selfNode(((tAttrData *)pNode))) ;
	else
	    pPad = Node_selfPad (pNode) ;

	if (pPad -> xDomTree != pDomTree -> xNdx)
	    NodePad_clonePad (pDomTree, pPad) ;
	}
    
    return pPad ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_appendChild                                                         */
/*                                                                          */
/* Append a child node to a parent node                                     */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tNode Node_appendChild (/*in*/  tDomTree *   pDomTree,
			/*in*/	tNodeType	 nType,
			/*in*/	int              bForceAttrValue,
			/*in*/	const char *	 sText,
			/*in*/	int		 nTextLen,
			/*in*/	tNode 		 xParent,
			/*in*/	int		 nLevel,
			/*in*/	int		 nLinenumber)

    {
    tNodeData *	pParent;
    tNodePad *  pPad ;

    pPad    = Node_condClonePad (pDomTree, xParent) ;
    pParent = Node_self (pDomTree, xParent) ;
    
    if (nType == ntypAttr)
	{	    
	tNodePad *   pPad = Node_selfPad (pParent) ;
	struct tAttrData *  pNew  ;
	tIndex		    xNdx ;

	pNew = ((struct tAttrData * )(pParent + 1)) + pParent -> numAttr ;
	
	if (((tUInt8 *)pNew) - ((tUInt8 *)pPad) != pPad -> nFill)
	    { /* not last child in pad -> move following node to new pad */
	    NodePad_splitPad (pDomTree, pPad, (tNodeData *)pNew, 0) ;
	    }
	if (((tUInt8 *)pNew) - ((tUInt8 *)pPad) + sizeof (struct tAttrData) > pPad -> nMax)
	    { /* pad full -> move into new pad */
	    pPad = NodePad_splitPad (pDomTree, pPad, pParent, ((tUInt8 *)pNew) - ((tUInt8 *)pParent) + sizeof (struct tAttrData)) ;
	    pParent = NodePad_selfFirstChild (pDomTree, pPad) ;
	    pNew = ((struct tAttrData * )(pParent + 1)) + pParent -> numAttr ;
	    }

        xNdx = ArrayAdd (&pDomTree -> pLookup, 1) ;
	pDomTree -> pLookup[xNdx] = (struct tNodeData *)pNew ;

	pNew -> xName  = sText?String2NdxNoInc (sText, nTextLen):nTextLen ;
	NdxStringRefcntInc (pNew -> xName) ;
	pNew -> xValue = 0 ;
	pNew -> bFlags = aflgOK ;
	pNew -> nType = nType ;
	pNew -> xNdx	= xNdx ;
	pNew -> nNodeOffset = ((tUInt8 *)pNew) - ((tUInt8 *)pParent) ;
        pParent -> numAttr++ ;
	pPad -> nFill += sizeof (struct tAttrData) ;
	numAttr++ ;

	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (pCurrReq, "[%d]PARSE: AddNode: +%02d %*s Attribut parent=%d node=%d type=%d text=%*.*s (#%d)\n", 
	                             pCurrReq -> nPid, nLevel, nLevel * 2, "", xParent, xNdx, nType, nTextLen, nTextLen, sText?sText:Ndx2String (nTextLen), sText?String2NdxNoInc (sText, nTextLen):nTextLen) ; 

	return xNdx ;
	}
    
    if ((bForceAttrValue || nType == ntypAttrValue) && (pParent -> nType != ntypAttr || (pParent -> bFlags & aflgAttrChilds) == 0))
	{	    
	struct tAttrData * pNew = (struct tAttrData * )pParent ; 
        int bAddChild = 0 ;
	
	if (pNew -> nType != ntypAttr)
            {   
	    if (nType == ntypAttrValue)
		{
		int i ;

		for (i = 0; i < nTextLen; i++)
		    {
		    if (!isspace (sText[i]))
			break ;
		    }
		
		if (i == nTextLen)
		    return 1 ; /* do not add only spaces as cdata inside a tag */
		}

	    pNew = ((tAttrData *)(pParent + 1)) + pParent -> numAttr - 1 ;
	    if (pParent -> numAttr == 0 || pNew -> xName != xNoName || bForceAttrValue > 1)
		{
		if (!(xParent = Node_appendChild (pDomTree, ntypAttr, 0, NULL, xNoName, xParent, nLevel, nLinenumber)))
		    return 0 ;
		nLevel++ ;
		pNew = (struct tAttrData * )pDomTree -> pLookup[xParent] ; 
                }
	    else
		{
		xParent = pNew -> xNdx ;
		bAddChild = 1 ;
		nLevel++ ;
		}
	    }
        
        
        if (!bAddChild && !bForceAttrValue)
	    {
	    pNew -> xValue  = sText?String2NdxNoInc (sText, nTextLen):nTextLen ;
	    NdxStringRefcntInc (pNew -> xValue) ;
	    if (pCurrReq -> bDebug & dbgParse)
		lprintf (pCurrReq, "[%d]PARSE: AddNode: +%02d %*s AttributValue parent=%d node=%d type=%d text=%*.*s (#%d)\n", pCurrReq -> nPid, nLevel, nLevel * 2, "", xParent, pNew -> xNdx, nType, nTextLen, nTextLen, sText, sText?String2NdxNoInc (sText,
 nTextLen):-1) ; 
	    pNew -> bFlags |= aflgAttrValue ;

	    return xParent ;
	    }
        else
    	    pParent = (tNodeData *)pNew ;

        }

    
	{
	struct tNodePad *   pChilds ;
	struct tNodeData *  pNew    ;
	tIndex		    xNdx ;

	if (pParent && pParent -> nType == ntypAttr)
	    {
	    tIndex xOldValue = 0 ;

            if (((tAttrData *)pParent) -> bFlags & aflgAttrChilds)
                pChilds = NodePad_self (pDomTree, ((tAttrData *)pParent) -> xValue) ;
            else
                pChilds = NewPad (pDomTree, xParent, 0)  ;
	    /* lprintf (pCurrReq, "[%d]PARSE: NewPad NewAttrChild: %d -> %d\n", pCurrReq -> nPid, xParent, pChilds -> xNdx) ; */

            if (((tAttrData *)pParent) -> bFlags & aflgAttrValue)
                {
                xOldValue = ((tAttrData *)pParent) -> xValue ;
                ((tAttrData *)pParent) -> bFlags &= ~aflgAttrValue ;
                }

            
            ((tAttrData *)pParent) -> xValue = pChilds -> xNdx ;
            ((tAttrData *)pParent) -> bFlags |= aflgAttrChilds ;
            if (xOldValue)
                NodePad_appendChild (pDomTree, &pChilds, ntypAttrValue, xOldValue, nLinenumber) ;

	    }
	else if (!pParent || !pParent -> xChilds)
	    {
	    pChilds = NewPad (pDomTree, xParent, 0)  ;
	    /* lprintf (pCurrReq, "[%d]PARSE: NewPad NewChild: %d -> %d\n", pCurrReq -> nPid, xParent, pChilds -> xNdx) ; */

	    if (pParent)
		pParent -> xChilds = pChilds -> xNdx ;
	    }
	else 
	    {
	    pChilds = NodePad_self (pDomTree, pParent -> xChilds) ;
	    }

        pNew = NodePad_appendChild (pDomTree, &pChilds, nType, String2Ndx (sText, nTextLen), nLinenumber) ;
	
	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (pCurrReq, "[%d]PARSE: AddNode: +%02d %*s Element parent=%d node=%d type=%d text=%*.*s (#%d)\n", pCurrReq -> nPid, nLevel, nLevel * 2, "", xParent, pNew -> xNdx, nType, nTextLen, nTextLen, sText, 
	                                                                        sText?String2NdxNoInc (sText, nTextLen):-1) ; 

	return pNew -> xNdx ;
	}
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_removeChild                                                         */
/*                                                                          */
/* Remove a child node                                                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tNode Node_removeChild (/*in*/ tDomTree *   pDomTree,
			/*in*/ tNode		xNode,
			/*in*/ tNode		xChild)

    {
    struct tNodeData *	pChild  = Node_self (pDomTree, xChild) ;
    
    pChild -> bFlags = nflgDeleted ;

    if (pChild -> nType != ntypAttr)
        {
        struct tNodePad *	pPad	= (struct tNodePad * )(((tUInt8 *)pChild) - pChild -> nPadOffset) ;

        if (xNode != -1 && pPad -> xParent != xNode)
	    return 0 ;

        /* pPad   -> numChilds-- ; */
        }

    return xChild ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_removeChild                                                         */
/*                                                                          */
/* Remove a child node                                                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */



struct tNodeData *  Node_selfRemoveChild (/*in*/  tDomTree *    pDomTree,
					  /*in*/ tNode		    xNode,
					  /*in*/ struct tNodeData * pChild)

    {
    tNodePad *	pPad	=   Node_condClonePad (pDomTree, pChild -> xNdx) ; 

    if (xNode != -1 && pPad -> xParent != xNode)
	return 0 ;

    /* relookup in case Pad has cloned */
    pChild = Node_self (pDomTree, pChild -> xNdx) ;
    pChild -> bFlags = nflgDeleted ;
    /* pPad   -> numChilds-- ; */

    return pChild ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_replaceChild                                                        */
/*                                                                          */
/* Replace child node                                                       */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNodeData * Node_selfCloneNode (/*in*/ tDomTree *      pDomTree,
				/*in*/ tNodeData *     pNode,
				/*in*/  int	       bDeep)

    {
    int	        len  = sizeof (tNodeData) + (bDeep == -1?0:pNode -> numAttr * sizeof (tAttrData)) ; 
    tNode       xParent = xNode_selfParentNode (pNode) ;
    tNodePad *  pPad = NewPad (pDomTree, xParent, len) ;
    tNodeData * pNew = NodePad_selfFirstChild (pDomTree,pPad) ;
    tNode	xNewNode ;

    memcpy (pNew, pNode, len) ;

    xNewNode = ArrayAdd (&pDomTree -> pLookup, 1) ;
    pDomTree -> pLookup[xNewNode] = pNew ;
    pNew -> xNdx	= xNewNode ;
    pNew -> nPadOffset = pPad -> nFill ;
    if (pNew -> nText)
	NdxStringRefcntInc (pNew -> nText) ;

    pPad -> numChilds++ ;
    pPad -> nFill += len ;
    numNodes++ ;

    if (bDeep == -1)
	pNew -> numAttr = 0 ;
    else
	{
	tAttrData * pAttr = (tAttrData * )(pNew + 1) ;
	int         n     = pNew -> numAttr ;

	while (n)
	    {
	    xNewNode = ArrayAdd (&pDomTree -> pLookup, 1) ;
	    pDomTree -> pLookup[xNewNode] = pAttr ;
	    pAttr -> xNdx = xNewNode ;
	    if (pAttr -> xName)
		NdxStringRefcntInc (pAttr -> xName) ;
	    if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
		NdxStringRefcntInc (pAttr -> xValue) ;
	    n-- ;
	    pAttr++ ;
	    }
	}
    if (bDeep < 1)
	pNew -> xChilds = 0 ;

    return pNew ;
    }
                                



tNode Node_cloneNode (/*in*/ tDomTree *      pDomTree,
		    /*in*/ tNode 	    xNode,
		    /*in*/  int		    bDeep)

    {
    tNodeData * pNew = Node_selfCloneNode (pDomTree, Node_self (pDomTree, xNode), bDeep) ;

    if (pNew)
	return pNew -> xNdx ;

    return 0 ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_replaceChild                                                        */
/*                                                                          */
/* Replaces the node pointed by pOldChildDomTree/xOldChild, with the node   */
/* pointed by pDomTree/xNode                                                */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tNode Node_replaceChildWithNode (/*in*/ tDomTree *      pDomTree,
				 /*in*/ tNode		xNode,
                                 /*in*/ tDomTree *      pOldChildDomTree,
				 /*in*/ tNode		xOldChild)

    {
    int		bFlags = nflgModified | nflgReturn ;
    tNode	xOrgChild  = xOldChild ;
    tNodeData *	pNode      = Node_self (pDomTree, xNode) ;
    tNodeData *	pOldChild  ;
    
    Node_condClonePad (pOldChildDomTree, xOldChild) ; 
    pOldChild  = Node_self (pOldChildDomTree, xOldChild) ;

    if (pOldChild -> bFlags & nflgModified) 
	{
	pOldChild -> bFlags |= bFlags ;
	pOldChild = Node_selfCloneNode (pOldChildDomTree, pNode, 1) ;
	xOldChild = pOldChild -> xNdx ;
	}
    else
	{
	int	len  = sizeof (tNodeData) + pNode -> numAttr * sizeof (tAttrData) ; 
	int     numAttr = pOldChild -> numAttr ;
	int		nOffset ;
	tAttrData * pAttr  ;
	int         n      ;

	if (pNode -> numAttr > pOldChild -> numAttr)
	    {
	    tAttrData *  pNew = ((struct tAttrData * )(pOldChild + 1)) + pOldChild -> numAttr ;
	    tNodePad  *  pPad = Node_selfPad (pOldChild) ;


	    if (((tUInt8 *)pNew) - ((tUInt8 *)pPad) != pPad -> nFill)
		{ /* not last child in pad -> move following node to new pad */
		NodePad_splitPad (pOldChildDomTree, pPad, (tNodeData *)pNew, 0) ;
		}

	    if (((tUInt8 *)pOldChild) - ((tUInt8 *)pPad) + len > pPad -> nMax)
		{ /* pad full -> move into new pad */
		pPad = NodePad_splitPad (pOldChildDomTree, pPad, pOldChild, len) ;
		pOldChild = NodePad_selfFirstChild (pOldChildDomTree, pPad) ;
		}
		
	    }
	
	nOffset = pOldChild -> nPadOffset ;
	xOldChild = pOldChild -> xNdx ;
	
	if (pOldChild -> nText)
	    NdxStringFree (pOldChild -> nText) ;

	pAttr = ((struct tAttrData * )(pOldChild + 1))  ;
	n     = pOldChild -> numAttr ;

	while (n > 0)
	    {
	    if (pAttr -> xName)
		NdxStringFree (pAttr -> xName) ;
	    if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
		NdxStringFree  (pAttr -> xValue) ;
	    n-- ;
	    pAttr++ ;
	    }

	memcpy (pOldChild, pNode, len) ;

	if (pOldChild -> nText)
	    NdxStringRefcntInc (pOldChild -> nText) ;
	

	pOldChild -> nPadOffset = nOffset ;
	pOldChild -> xNdx       = xOldChild ;
	pOldChild -> bFlags |= bFlags ;

	pAttr = ((struct tAttrData * )(pOldChild + 1))  ;
	n     = pNode -> numAttr ;

	while (n > 0)
	    {
	    if (pAttr -> xName)
		NdxStringRefcntInc (pAttr -> xName) ;
	    if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
		NdxStringRefcntInc  (pAttr -> xValue) ;
	    n-- ;
	    pAttr++ ;
	    }
	
	pAttr = ((struct tAttrData * )(pOldChild + 1)) + pOldChild -> numAttr ;
	n     = numAttr - pNode -> numAttr ;

	while (n > 0)
	    {
	    pAttr -> bFlags = nflgDeleted ;
	    if (pAttr -> xName)
		NdxStringFree (pAttr -> xName) ;
	    if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
		NdxStringFree (pAttr -> xValue) ;
	    n-- ;
	    pAttr++ ;
	    }
	}

    if (pOldChild -> nType == ntypDocumentFraq)
	{
	tAttrData * pAttr = Element_selfSetAttribut (pOldChildDomTree, pOldChild, NULL, xDomTreeAttr, NULL, pDomTree -> xNdx, 0) ;
	pAttr -> bFlags = aflgOK ; /* reset string value flag */
	}
	
    DomTree_selfCheckpoint (pOldChildDomTree, xOrgChild, xOldChild) ;
    
    return xOldChild ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_replaceChild                                                        */
/*                                                                          */
/* Replace child node                                                       */
/*                                                                          */
/* ------------------------------------------------------------------------ */

tNode Node_replaceChildWithCDATA (/*in*/ tDomTree *	 pDomTree,
				  /*in*/ tNode		 xNode,
				  /*in*/ tNode		 xOldChild,
			          /*in*/ const char *	 sText,
				  /*in*/ int		 nTextLen,
				  /*in*/ int		 nEscMode,
				  /*in*/ int		 bFlags)

    {
    tNodeData *	pOldChild  ;
    tNode	xOrgChild  = xOldChild ;
    tStringIndex n ;
    
    Node_condClonePad (pDomTree, xOldChild) ; 
    pOldChild  = Node_self (pDomTree, xOldChild) ;

    lprintf (pCurrReq, "rp1--> SVs=%d  %s  DomTree Old=%d\n", sv_count, sText, Node_selfDomTree (pOldChild)) ;

    xCheckpointCache[nCheckpointCache++] = xOldChild ;

    if (pOldChild -> bFlags & nflgModified) 
	{
	pOldChild -> bFlags |= bFlags ;
	pOldChild = Node_selfCloneNode (pDomTree, pOldChild, 0) ;
	xOldChild = pOldChild -> xNdx ;
	}
    lprintf (pCurrReq, "rp2--> DomTree New=%d\n", Node_selfDomTree (pOldChild)) ;
    
    DomTree_selfCheckpoint (pDomTree, xOrgChild, xOldChild) ;
    
    xCheckpointCache[nCheckpointCache++] = xOldChild ;
    nCheckpointCache &= nCheckpointCacheMask ;
    xCheckpointCache[nCheckpointCache] = 0 ;

   
    if (nEscMode != -1)
	{
	pOldChild -> nType  = (nEscMode & 3)?ntypText:ntypCDATA ;
	pOldChild -> bFlags &= ~(nflgEscUrl + nflgEscChar) ;
	pOldChild -> bFlags |= (nEscMode ^ nflgEscChar) & (nflgEscUrl + nflgEscChar) ;
	}
    else
	pOldChild -> nType  = ntypCDATA ;

    n = pOldChild -> nText ;
    pOldChild -> nText = String2Ndx(sText, nTextLen) ;
    pOldChild -> xChilds = 0 ;
    pOldChild -> bFlags |= bFlags ;
    if (n)
	NdxStringFree (n) ;

    lprintf (pCurrReq, "rp<-- nText=%d sText=>%*.*s< nTextLen = %d  SVs=%d\n", pOldChild -> nText, nTextLen,nTextLen, sText,  nTextLen, sv_count) ;
    return xOldChild ;
    }



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_parentNode                                                          */
/*                                                                          */
/* Get parent node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNode Node_parentNode  (/*in*/  tDomTree *   pDomTree,
			/*in*/	tNode 		 xNode) 

    {
    struct tNodeData *	pNode = Node_self (pDomTree, xNode) ;
    struct tNodePad *	pPad	= (struct tNodePad * )(((tUInt8 *)pNode) - pNode -> nPadOffset) ;
    return pPad -> xParent ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_selfParentNode                                                      */
/*                                                                          */
/* Get parent node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNodeData * Node_selfParentNode  (/*in*/  tDomTree *	pDomTree,
				  /*in*/  tNodeData *	pNode) 

    {
    struct tNodePad *	pPad	= (struct tNodePad * )(((tUInt8 *)pNode) - pNode -> nPadOffset) ;
    return Node_self (pDomTree, pPad -> xParent) ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_selfParentNode                                                      */
/*                                                                          */
/* Get parent node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNode xNode_selfParentNode  (/*in*/  tNodeData *	pNode) 

    {
    tNodePad *	pPad	= (struct tNodePad * )(((tUInt8 *)pNode) - pNode -> nPadOffset) ;
    return pPad -> xParent ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_firstChild                                                          */
/*                                                                          */
/* Get first child node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNode Node_firstChild  (/*in*/  tDomTree *   pDomTree,
			/*in*/	tNode 		 xNode) 

    {
    struct tNodePad *	pPad = NodePad_self (pDomTree, Node_self (pDomTree, xNode) -> xChilds) ;
    tNodeData * pNode ;
    if (pPad == NULL)
	return 0 ;
    
    pNode = ((tNodeData *)(pPad + 1)) ;
    if (!pNode -> bFlags)
	pNode = Node_selfNextSibling (pDomTree, pNode) ;

    return pNode && pNode -> bFlags?pNode -> xNdx:0 ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_selfFirstChild                                                          */
/*                                                                          */
/* Get first child node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNodeData * Node_selfFirstChild  (/*in*/  tDomTree *   pDomTree,
				  /*in*/  tNodeData *  pNode) 

    {
    struct tNodePad *	pPad = NodePad_self (pDomTree, pNode -> xChilds) ;
    tNodeData * pChildNode ;
    if (pPad == NULL)
	return 0 ;
    
    pChildNode = ((tNodeData *)(pPad + 1)) ;
    if (!pChildNode -> bFlags)
	pChildNode = Node_selfNextSibling (pDomTree, pChildNode) ;

    return pChildNode ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_selfLastChild                                                       */
/*                                                                          */
/* Get last child node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNodeData * Node_selfLastChild   (/*in*/  tDomTree *   pDomTree,
				  /*in*/  tNodeData *  pNode) 

    {
    struct tNodePad *	pPad = NodePad_self (pDomTree, pNode -> xChilds) ;
    tNodeData * pChildNode ;
    tNodeData * pLastChild ;
    if (pPad == NULL)
	return NULL ;
    
    if (pPad -> xLast && pPad -> xLast != pPad -> xNdx)
	pPad = NodePad_self (pDomTree, pPad -> xLast) ;
    
    pLastChild = NULL ;
    do
	{
	pChildNode = ((tNodeData *)(pPad + 1)) ;

	while (pChildNode)
	    {
	    pLastChild = pChildNode ;
	    pChildNode = Node_selfNextSibling (pDomTree, pChildNode) ;
	    }

	if (!pLastChild)
	    {
	    if (pPad -> xPrev)
		pPad = NodePad_self (pDomTree, pPad -> xPrev) ;
	    else
		return NULL ;
	    }
	}
    while (!pLastChild) ;

    return pLastChild ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_selfNthChild (pNode, nChildNo) ;                                    */
/*                                                                          */
/* Get first child node							    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


struct tNodeData * Node_selfNthChild (/*in*/  tDomTree *    pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ int		nChildNo) 

    {
    struct tNodePad *	pPad = pDomTree -> pLookup[pNode -> xChilds] ;
    struct tNodeData *  pChildNode ;

    if (pPad == NULL)
	return 0 ;

    while (nChildNo >= pPad -> numChilds && pPad -> xNext)
	{
	nChildNo -= pPad -> numChilds ;
	pPad = pDomTree -> pLookup[pPad -> xNext] ;
	}

    if (nChildNo < 0)
	return NULL ;

    
    pChildNode = ((struct tNodeData *)(pPad + 1)) ;
    while (nChildNo > 0)
	{
	if (pChildNode -> bFlags)
	    nChildNo-- ;
	pChildNode = (struct tNodeData *)(((tUInt8 *)pChildNode) + sizeof (struct tNodeData) + sizeof (struct tAttrData) * pChildNode -> numAttr) ;
	}

    
    return pChildNode ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_nextSibling (xNode) ;                                               */
/*                                                                          */
/* Get next sibling node						    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNodeData * Node_selfNextSibling (/*in*/ tDomTree *  pDomTree,
			          /*in*/ tNodeData * pNode) 

    {
    struct tNodePad *	pPad	= (struct tNodePad * )(((tUInt8 *)pNode) - pNode -> nPadOffset) ;

    do
	{
	int nOffset = sizeof (struct tNodeData) + sizeof (struct tAttrData) * pNode -> numAttr ;
	if (pPad -> nFill <= pNode -> nPadOffset + nOffset || pPad -> numChilds == 0) 
	    { /* next pad */
	    if (!pPad -> xNext)
		return 0 ;

	    pPad  = pDomTree -> pLookup[pPad -> xNext] ;
	    pNode = ((struct tNodeData *)(pPad + 1))  ;
	    }
	else
	    pNode = (struct tNodeData *)(((tUInt8 *)pNode) + nOffset) ;
	}
    while (!pNode -> bFlags) ;
        
    return pNode  ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_nextSibling (xNode) ;                                               */
/*                                                                          */
/* Get next sibling node						    */
/*                                                                          */
/* ------------------------------------------------------------------------ */


tNode Node_nextSibling (/*in*/ tDomTree *  pDomTree,
			/*in*/ tNode       xNode) 


    {
    tNodeData * pNode = Node_self (pDomTree, xNode) ;
    tNodeData * pNext = Node_selfNextSibling(pDomTree, pNode) ;

    return pNext?pNext -> xNdx:0 ;
    }

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Node_toString                                                            */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tNodeData * Node_toString2 (/*in*/  tDomTree *     pDomTree,
		    /*i/o*/ register req *  r,
		    /*in*/  tNode	    xNode,
                    /*in*/  int *           pOrderNdx)


    {
    tNode xFirstNode = xNode ;
    tNode xNextNode ;
    int   nOrderNdx = *pOrderNdx ;
    int   nOrderIndexNode = -1 ;
    int   bCheckpointFound = 0  ;
    tNodeData * pNextNode ;
    tNodeData * pSavedNode = NULL ;
    tNodeData * pSavedNodeFrom = NULL ;
    tNodeData * pLastStartTag = NULL ;
    tNodeData * pNode = Node_self (pDomTree, xNode) ;
    tNodeData * pFirstNode = pNode ;
    
    
    
    if (pNode -> nType == ntypDocumentFraq)
	pDomTree = DomTree_self (Element_selfGetAttribut (pDomTree, pNode, NULL, xDomTreeAttr) -> xValue) ;
    
    pNode = Node_selfFirstChild (pDomTree, pNode) ;

    while (pNode)
	{
	if (pNode -> bFlags & nflgCheckpoint)
	    { /* see how the control flow contiounes */
	    if ((xNextNode = pDomTree -> pOrder[nOrderNdx].xFromNode) != pNode -> xNdx && xNextNode)	
		{
		tNodeData * pNewNode ;
                tNodeData * pNewParent ;
                tNodeData * pParent ;
                
                if (pCurrReq -> bDebug & dbgParse)
		    lprintf (r, "[%d]toString: Skip Node=%d to Node=%d\n", r -> nPid, pNode -> xNdx, xNextNode) ; 
		bCheckpointFound = 1 ;
                pSavedNode = NULL ;
		pNewNode      = Node_self (pDomTree, xNextNode) ;
		pNewParent    = Node_selfParentNode (pDomTree, pNewNode) ;
		if (pParent       = pLastStartTag)
		    {
		    while (pNewParent != pParent)
			{
			if (pParent)
			    {
			    oputs (r, "</") ;
			    oputs (r, Node_selfNodeName (pParent)) ;
			    oputc (r, '>') ;
			    }
			else
			    {
			    lprintf (r, "[%d]WARN: unstructured jump\n", r -> nPid) ;
			    break ;
			    }
			pParent = Node_selfParentNode (pDomTree, pParent) ;
			}
		    }
		pLastStartTag = NULL ;
		pNode = pNewNode ;
		}
	    nOrderNdx++ ;
	    }

	if (pNode -> bFlags & nflgReturn)
	    { /* we should use another node to replace this one */
	    tDomTreeOrder * pOrder = pDomTree -> pOrder ;
	    int             n      = ArrayGetSize (pOrder) ;
	    int             i      = nOrderNdx ;

	    while (pOrder[i].xFromNode != pNode -> xNdx && i < n)  
		i++ ;
	    
	    if (i < n)
		{		    
		if ((xNextNode = pOrder[i].xToNode) != pNode -> xNdx && xNextNode)	
		    {
		    if (pCurrReq -> bDebug & dbgParse)
			lprintf (r, "[%d]toString: Replace Node=%d with Node=%d\n", r -> nPid, pNode -> xNdx, xNextNode) ; 
		    pSavedNode = pNode ;    
    		    pNode  = Node_self (pDomTree, xNextNode) ;
		    pSavedNodeFrom = pNode ;    
		    }
		if (i == nOrderNdx)
		    nOrderNdx++, nOrderIndexNode = -1 ;
		else
		    nOrderIndexNode = i ;
		}
	    else
		{
		mydie ("Internal Error: Orderindex out of range") ;
		}
	    }


	if (pCurrReq -> bDebug & dbgParse)
	    lprintf (r, "[%d]toString: Node=%d type=%d flags=%x text=>%s<= (#%d) SVs=%d\n", r -> nPid, pNode -> xNdx, pNode -> nType,  pNode -> bFlags, Ndx2String (pNode -> nText), pNode -> nText, sv_count) ; 

	if (pNode -> bFlags & nflgIgnore)
            ;
	else if (pNode -> nType == ntypDocumentFraq)
	    {
	    Node_toString (pDomTree, r, pNode -> xNdx) ; 
	    oputs (r, "\r\n") ;
	    }
        else if (pNode -> nType == ntypTag || pNode -> nType == ntypStartTag)
	    {
	    int n = pNode -> numAttr ;
	    struct tAttrData * pAttr = (struct tAttrData *)(pNode + 1) ;

	    if (pNode -> nType == ntypStartTag)
		pLastStartTag = pNode ;

	    oputc (r, '<') ;
	    oputs (r, Node_selfNodeName (pNode)) ;
	    
	    while (n--)
		{
		if (pAttr -> bFlags)
		    {
		    char * s ;
		    int    l ;
		    oputc (r, ' ') ;
		    if (pAttr -> xName != xNoName)
			{
			Ndx2StringLen (pAttr -> xName,s,l) ;
			owrite (r, s, l);
			}

		    if (pAttr -> xValue)
			{
			if (pAttr -> xName != xNoName)
			    oputs (r, "=\"") ;

			if (pAttr -> bFlags & aflgAttrChilds)
			    {
			    tAttrData * pAttrNode = (tAttrData * )Node_toString2 (pDomTree, r, pAttr -> xNdx, &nOrderNdx) ;
                            if (pAttrNode && pAttrNode != pAttr)
                                {
                                pAttr = pAttrNode ;
                                pNode = Attr_selfNode(pAttr) ;
                                n = pNode -> numAttr - Attr_selfAttrNum (pAttr) - 1 ;
                                if (n < 0)
                                    n = 0 ;
                                }
			    }
			else
			    {
			    Ndx2StringLen (pAttr -> xValue, s, l) ;
			    while (isspace (*s) && l > 0)
				s++, l-- ;
			    owrite (r, s, l) ;
			    }
			if (pAttr -> xName != xNoName)
			    oputc (r, '"') ;
			}
		    }
		pAttr++ ;
		}
	    oputc (r, '>') ;
	    }
	else if (pNode -> nType == ntypText)
	    {
	    char * s ;
	    int    l ;
	    Ndx2StringLen (pNode -> nText,s,l) ;
	    OutputEscape (r, s, l, (pNode -> bFlags & nflgEscUrl)?Char2Url:Char2Html, (pNode -> bFlags & nflgEscChar)?'\\':0) ;
	    }
	else 
	    {
	    char * s ;
	    int    l ;
	    Ndx2StringLen (pNode -> nText,s,l) ;
	    owrite (r, s, l);
	    }

	if (nOrderNdx == nOrderIndexNode)
	    nOrderNdx++ ;

	if (pNode -> nType == ntypDocumentFraq)
	    pNextNode = NULL ;
	else
	    pNextNode = Node_selfFirstChild (pDomTree, pNode) ;
	if (pNextNode == NULL)
	    {
	    if (pSavedNode && pSavedNodeFrom == pNode)
		{
		pNode = pSavedNode ;
		pSavedNode = NULL ;
		}

	
	    pNextNode  = Node_selfNextSibling (pDomTree, pNode) ;
	    while (pNextNode == NULL)
		{
		pNextNode = Node_selfParentNode (pDomTree, pNode) ;
		if (pNextNode == NULL || pNextNode  == pFirstNode || pNextNode -> nType == ntypAttr)
                    {
                    *pOrderNdx = nOrderNdx ;
                    return bCheckpointFound?pNextNode:NULL ;
                    }
		    
		if (pSavedNode && pSavedNodeFrom == pNextNode)
		    {
		    pNextNode = pSavedNode ;
		    pSavedNode = NULL ;
		    }
		pNode = pNextNode ;
		pNextNode  = Node_selfNextSibling (pDomTree, pNextNode) ;
		if (pNode -> nType == ntypStartTag && (pNode -> bFlags & nflgIgnore) == 0 &&
		    (pNextNode == NULL || (pNextNode -> bFlags & nflgCheckpoint) == 0 || pDomTree -> pOrder[nOrderNdx].xFromNode == pNextNode -> xNdx))
		    {
		    oputs (r, "</") ;
		    oputs (r, Node_selfNodeName (pNode)) ;
		    oputc (r, '>') ;
		    pLastStartTag = Node_selfParentNode (pDomTree, pNextNode?pNextNode:pNode) ;
		    }
		}
	    }
	pNode = pNextNode ;
	}

    *pOrderNdx = nOrderNdx ;

    return NULL ;    
    }

void Node_toString (/*in*/  tDomTree *   pDomTree,
		    /*i/o*/ register req * r,
		    /*in*/  tNode	    xNode)


    {
    int   nOrderNdx = 0 ;

    Node_toString2 (pDomTree, r, xNode, &nOrderNdx) ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* NodeList_toString                                                            */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */


void NodeList_toString (/*in*/  tDomTree *   pDomTree,
			/*in*/ tNode	    xNode)


    {
    xNode = Node_firstChild (pDomTree, xNode) ;

    

    }

    
/* ------------------------------------------------------------------------

interface Element : Node {
  readonly attribute DOMString        tagName;
  DOMString          getAttribute(in DOMString name);
  void               setAttribute(in DOMString name, 
                                  in DOMString value)
                                        raises(DOMException);
  void               removeAttribute(in DOMString name)
                                        raises(DOMException);
  Attr               getAttributeNode(in DOMString name);
  Attr               setAttributeNode(in Attr newAttr)
                                        raises(DOMException);
  Attr               removeAttributeNode(in Attr oldAttr)
                                        raises(DOMException);
  NodeList           getElementsByTagName(in DOMString name);
  // Introduced in DOM Level 2:
  DOMString          getAttributeNS(in DOMString namespaceURI, 
                                    in DOMString localName);
  // Introduced in DOM Level 2:
  void               setAttributeNS(in DOMString namespaceURI, 
                                    in DOMString qualifiedName, 
                                    in DOMString value)
                                        raises(DOMException);
  // Introduced in DOM Level 2:
  void               removeAttributeNS(in DOMString namespaceURI, 
                                       in DOMString localName)
                                        raises(DOMException);
  // Introduced in DOM Level 2:
  Attr               getAttributeNodeNS(in DOMString namespaceURI, 
                                        in DOMString localName);
  // Introduced in DOM Level 2:
  Attr               setAttributeNodeNS(in Attr newAttr)
                                        raises(DOMException);
  // Introduced in DOM Level 2:
  NodeList           getElementsByTagNameNS(in DOMString namespaceURI, 
                                            in DOMString localName);
  // Introduced in DOM Level 2:
  boolean            hasAttribute(in DOMString name);
  // Introduced in DOM Level 2:
  boolean            hasAttributeNS(in DOMString namespaceURI, 
                                    in DOMString localName);
};


------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Element_selfGetAttribut                                                  */
/*                                                                          */
/* Get attribute value of Element by name                                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tAttrData *  Element_selfGetAttribut (/*in*/ tDomTree *	        pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen) 

    {
    int nAttrName = sAttrName?String2NdxNoInc (sAttrName, nAttrNameLen):nAttrNameLen ;
    struct tAttrData * pAttr = (struct tAttrData * )(pNode + 1) ;
    int  n = pNode -> numAttr ;

    while (n > 0 && (nAttrName != pAttr -> xName || !pAttr -> bFlags))
	{
	n-- ;
	pAttr++ ;
	}

    if (n)
	return pAttr ;

    return NULL ;
    }



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Element_selfGetNthAttribut                                               */
/*                                                                          */
/* Get attribute value of Element by index                                  */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tAttrData *  Element_selfGetNthAttribut (/*in*/ tDomTree *	   pDomTree,
			  	         /*in*/ struct tNodeData * pNode,
				         /*in*/ int                n)

    {
    struct tAttrData * pAttr = (struct tAttrData * )(pNode + 1) ;
    int  num = pNode -> numAttr ;

    if (n < 0 || n >= num)
        return NULL ;

    return pAttr + n ;
    }




/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Element_selfSetAttribut                                                  */
/*                                                                          */
/* Set attribute value of Element by name                                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tAttrData *  Element_selfSetAttribut (/*in*/ tDomTree *	        pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen,
				      /*in*/ const char *       sNewValue, 
				      /*in*/ int		nNewValueLen,
				      /*in*/ int		bClone)

    {
    tAttrData * pAttr ;
    tNode xAttr ;
    tNodeData * pNewNode ;

    Node_condClonePad (pDomTree, pNode -> xNdx) ;
    pNode = Node_self (pDomTree, pNode -> xNdx) ;
    pAttr = Element_selfGetAttribut (pDomTree, pNode, sAttrName, nAttrNameLen) ;

    if (pAttr)
	{
	tIndex xValue = sNewValue?String2NdxNoInc (sNewValue, nNewValueLen):nNewValueLen ;
	NdxStringRefcntInc (xValue) ;

	if (bClone)
	    {
	    if (pAttr -> xValue != xValue && (pNode -> bFlags & nflgModified))
		{
		pNewNode = Node_selfCloneNode (pDomTree, pNode, 1) ;
		pAttr = Element_selfGetAttribut (pDomTree, pNewNode, sAttrName, nAttrNameLen) ;
		
		DomTree_selfCheckpoint (pDomTree, pNode -> xNdx, pNewNode -> xNdx) ;
		}
	    else
		DomTree_selfCheckpoint (pDomTree, pNode -> xNdx, pNode -> xNdx) ;

	    pNode -> bFlags |= nflgReturn | nflgModified ;
	    }

	if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
	    NdxStringFree (pAttr -> xValue) ;

	pAttr -> xValue = xValue ;
	return pAttr ;
	}

    if (pNode -> bFlags & nflgModified)
	pNewNode = Node_selfCloneNode (pDomTree, pNode, 1) ;
    else
	pNewNode = pNode ;

    if (bClone)
	{
	pNode -> bFlags |= nflgReturn | nflgModified ;
	DomTree_selfCheckpoint (pDomTree, pNode -> xNdx, pNewNode -> xNdx) ;
	}
    
    xAttr = Node_appendChild (pDomTree, ntypAttr, 0, sAttrName, nAttrNameLen, pNewNode -> xNdx, 0, pNewNode -> nLinenumber) ;
    Node_appendChild (pDomTree, ntypAttrValue, 0, sNewValue, nNewValueLen, xAttr, 0, pNewNode -> nLinenumber) ;
    return (tAttrData *)Node_self(pDomTree, xAttr) ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Element_selfRemoveAttribut                                               */
/*                                                                          */
/* Remove attribute of Element by name                                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */



tAttrData *  Element_selfRemoveAttribut (/*in*/ tDomTree *	        pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen,
				      /*in*/ int		bClone)

    {
    tAttrData * pAttr ;
    tNodeData * pNewNode ;
    
    Node_condClonePad (pDomTree, pNode -> xNdx) ;
    pNode = Node_self (pDomTree, pNode -> xNdx) ;

    pAttr = Element_selfGetAttribut (pDomTree, pNode, sAttrName, nAttrNameLen) ;
    if (bClone)
	{
	if (pAttr != NULL && (pNode -> bFlags & nflgModified))
	    {
	    pNode -> bFlags |= nflgReturn | nflgModified ;

	    pNewNode = Node_selfCloneNode (pDomTree, pNode, 1) ;
	    pAttr = Element_selfGetAttribut (pDomTree, pNewNode, sAttrName, nAttrNameLen) ;
    
	    DomTree_selfCheckpoint (pDomTree, pNode -> xNdx, pNewNode -> xNdx) ;
	    }
	else
	    {
	    pNode -> bFlags |= nflgReturn | nflgModified ;

	    DomTree_selfCheckpoint (pDomTree, pNode -> xNdx, pNode -> xNdx) ;
	    }
	}

    if (pAttr)
	{
	pAttr -> bFlags = 0 ;
	if (pAttr -> xName)
	    NdxStringFree (pAttr -> xName) ;
	if (pAttr -> xValue && (pAttr -> bFlags & aflgAttrValue))
	    NdxStringFree (pAttr -> xValue) ;
	}

    /*
    int    nCopyAttr ;

    if (pAttr)
	{
	nAttr = pAttr - ((struct tAttrData * )(pNode + 1)) ;
	pNode -> numAttr-- ;
	nCopyAttr = pNode -> numAttr - nAttr  ;
	if (nCopyAttr)
	    {
	    memcpy (pAttr, pAttr + 1, nCopyAttr * sizeof (tAttrData)) ;
	    while (nCopyAttr--)
		{
		pDomTree -> pLookup[pAttr -> xNdx] = pAttr ;
		pAttr++ ;
		}
	    }
	}
    */

    return pAttr ;
    }


/* ------------------------------------------------------------------------


interface Attr : Node {
  readonly attribute DOMString        name;
  readonly attribute boolean          specified;
           attribute DOMString        value;
                                        // raises(DOMException) on setting

  // Introduced in DOM Level 2:
  readonly attribute Element          ownerElement;
};

------------------------------------------------------------------------ */



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Attr_selfValue							    */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* ------------------------------------------------------------------------ */



char *       Attr_selfValue (/*in*/ tDomTree *	        pDomTree,
			     /*in*/ struct tAttrData * pAttr,
			    /*out*/ char * *    ppAttr) 

    {
    struct tNodeData * pNode ;
    
    if (!pAttr)
	return NULL ;

    if (!(pAttr -> bFlags & aflgAttrChilds))
	return Ndx2String (pAttr -> xValue) ;

    pNode = NodePad_selfFirstChild (pDomTree, NodePad_self(pDomTree, pAttr -> xValue)) ;
    
    StringNew (ppAttr, 512) ;

    while (pNode)
	{
	char * s ;
	int    l ;

	if (pNode -> bFlags & nflgReturn)
	    {
	    tIndex xNdx = pNode -> xNdx ;
	    tNodeData * pUseNode ;
	    int i = nCheckpointCache - 2 ;
	
	    i &= nCheckpointCacheMask ;
	    
	    while (xCheckpointCache[i] && xCheckpointCache[i] != xNdx)
		{
		i -= 2 ;
		i &= nCheckpointCacheMask ;
		}

	    if (xCheckpointCache[i])
		xNdx = xCheckpointCache[i+1] ;

	    pUseNode = Node_self (pDomTree, xNdx) ;
	    Ndx2StringLen (pUseNode -> nText,s,l) ;
	    }
	else	
	    Ndx2StringLen (pNode -> nText,s,l) ;
	StringAdd (ppAttr, s, l) ;
	pNode = Node_selfNextSibling (pDomTree, pNode) ;
	}

    return *ppAttr ;
    }
