/*###################################################################################
#
#   Embperl - Copyright (c) 1997-2000 Gerald Richter / ECOS
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.
#
#   THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#   WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
###################################################################################*/


#include "ep.h"
#include "epmacro.h"


typedef   unsigned char tCharMap [256/(sizeof(unsigned char)*8)]   ;

struct tToken
    {
    const char *	    sText ;	/* string of token (MUST be first item!) */
    const char *	    sName ;	/* name of token (only for description) */
    int			    nTextLen ;	/* len of string */
    const char *	    sEndText ;	/* string which ends the block */
    const char *	    sNodeName;	/* name of the node to create */
    int			    nNodeName ;	/* index in string table of node name */
    enum tNodeType	    nNodeType ;	/* type of the node that should be created */
    enum tNodeType	    nCDataType ;/* type for sub nodes that contains text */
    enum tNodeType	    nForceType ;/* force this type for sub nodes */
    int			    bUnescape ;	/* translate input?  */
    unsigned char *	    pContains ;	/* chars that could becontains in the string */
    struct tTokenTable *    pFollowedBy;/* table of tokens that can follow this one */
    struct tTokenTable *    pInside ;	/* table of tokens that can apear inside this one */
    struct tToken      *    pStartTag ;	/* token that contains definition for the start of the current token */
    struct tToken      *    pEndTag ;	/* token that contains definition for the end of the current token */
    } ;        

struct tTokenTable
    {
    tCharMap	    cStartChars ;   /* for every vaild start char there is one bit set */
    tCharMap	    cAllChars   ;   /* for every vaild char there is one bit set */
    struct tToken * pTokens ;	    /* table with all tokens */
    int             numTokens ;	    /* number of tokens in above table */
    int		    bLSearch ;	    /* when set perform a linear, instead of a binary search */
    struct tToken * pContainsToken ;/* pointer to the token that has a pContains defined (could be only one per table) */
    } ;


struct tTokenCmp
    {
    char *          pStart ;
    char *          pCurr ;
    int		nLen ;
    } ;



struct tTokenTable DefaultTokenTable ;


#define parse_malloc(a,b) malloc(b)

/* ------------------------------------------------------------------------ */
/* compare tokens                                                           */
/* ------------------------------------------------------------------------ */

static int CmpToken (/*in*/ const void *  p1,
                     /*in*/ const void *  p2)

    {
    return strcmp (*((const char * *)p1), *((const char * *)p2)) ;
    }
    
/* ------------------------------------------------------------------------ */
/* compare tokens for descending order                                      */
/* ------------------------------------------------------------------------ */

static int CmpTokenDesc (/*in*/ const void *  p1,
                     /*in*/ const void *  p2)

    {
    return strcmp (*((const char * *)p2), *((const char * *)p1)) ;
    }


	    
/* ------------------------------------------------------------------------ */
/*                                                                          */
/* CheckProcInfo                                                            */
/*                                                                          */
/* Check for processor informations                                         */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int CheckProcInfo      (/*i/o*/ register req * r,
			/*in*/  HV *           pHash,
			/*in*/  struct tToken * pToken)
			

    {	
    int		    rc ;
    HE *	    pEntry ;
    char *	    pKey ;
    SV * *	    ppSV ;
    SV * 	    pSVValue ;
    IV		    l	 ;
    HV *            pHVProcInfo ;
    
    ppSV = hv_fetch(pHash, "procinfo", sizeof ("procinfo") - 1, 0) ;  
    if (ppSV != NULL)
	{		
	if (*ppSV == NULL || !SvROK (*ppSV) || SvTYPE (SvRV (*ppSV)) != SVt_PVHV)
	    {
	    strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
	    sprintf (r -> errdat2, "%s => procinfo", pToken -> sText) ;
	    return rcNotHashRef ;
	    }

	pHVProcInfo = (HV *)SvRV (*ppSV) ;

	hv_iterinit (pHVProcInfo ) ;
	while (pEntry = hv_iternext (pHVProcInfo))
	    {
	    HV *   pProcInfoHash ;
        
	    pKey     = hv_iterkey (pEntry, &l) ;
	    pSVValue = hv_iterval (pHVProcInfo , pEntry) ;
        
	    if (pSVValue == NULL || !SvROK (pSVValue) || SvTYPE (SvRV (pSVValue)) != SVt_PVHV)
		{
		strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
		sprintf (r -> errdat2, "%s => procinfo", pToken -> sText) ;
		return rcNotHashRef ;
		}

	    if (strcmp (pKey, "embperl") == 0)
		embperl_CompileInitItem (r, (HV *)(SvRV (pSVValue)), pToken -> nNodeName, pToken -> nNodeType) ;
	    }
	}	    


    return ok ;
    }


	    
	    


    
/* ------------------------------------------------------------------------ */
/*                                                                          */
/* BuildSubTokenTable                                                       */
/*                                                                          */
/* Build the C token table out of a Perl Hash                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int BuildSubTokenTable (/*i/o*/ register req * r,
			/*in*/  HV *           pHash,
			/*in*/  const char *   pKey,
			/*in*/  const char *   pAttr,
			/*in*/  const char *   pDefEnd,
			/*out*/ struct tTokenTable * * pTokenTable)

                     
    {
    SV * *  ppSV ;
    int	    rc ;
    
    ppSV = hv_fetch(pHash, (char *)pAttr, strlen (pAttr), 0) ;  
    if (ppSV != NULL)
	{		
	struct tTokenTable * pNewTokenTable ;
	HV *                 pSubHash ;

	if (*ppSV == NULL || !SvROK (*ppSV) || SvTYPE (SvRV (*ppSV)) != SVt_PVHV)
	    {
	    strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
	    sprintf (r -> errdat2, "%s => %s", pKey, pAttr) ;
	    return rcNotHashRef ;
	    }
	
	pSubHash = (HV *)SvRV (*ppSV) ;
	if ((pNewTokenTable = (struct tTokenTable *)GetHashValueInt (pSubHash, "--cptr", 0)) == NULL)
	    {
	    if ((pNewTokenTable = parse_malloc (r, sizeof (struct tTokenTable))) == NULL)
		 return rcOutOfMemory ;

	    if (r -> bDebug & dbgBuildToken)
		lprintf (r, "[%d]TOKEN: -> %s\n", r -> nPid, pAttr) ; 
	    if ((rc = BuildTokenTable (r, pSubHash, pDefEnd, pNewTokenTable)))
		return rc ;    
	    if (r -> bDebug & dbgBuildToken)
		lprintf (r, "[%d]TOKEN: <- %s\n", r -> nPid, pAttr) ; 
	    
	    if (pNewTokenTable -> numTokens == 0)
		{
		strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
		sprintf (r -> errdat2, "%s => %s does not contain any tokens", pKey, pAttr) ;
		return rcNotFound ;
		}

	    hv_store(pSubHash, "--cptr", sizeof ("--cptr") - 1, newSViv ((IV)pNewTokenTable), 0) ;
	    }
	else
	    if (r -> bDebug & dbgBuildToken)
	        lprintf (r, "[%d]TOKEN: -> %s already build; numTokens=%d\n", r -> nPid, pAttr, pNewTokenTable->numTokens) ; 
	

	*pTokenTable = pNewTokenTable ;
	return  ok  ;
	}

    *pTokenTable = NULL ;
    return ok ;
    }
    

/* ------------------------------------------------------------------------ */
/*                                                                          */
/* BuildTokenTable                                                          */
/*                                                                          */
/* Build the C token table out of a Perl Hash                               */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int BuildTokenTable (/*i/o*/ register req *	  r,
                     /*in*/  HV *		  pTokenHash,
		     /*in*/  const char *         pDefEnd,
                     /*out*/ struct tTokenTable * pTokenTable)

                     
    {
    int		    rc ;
    SV *	    pToken ;
    HE *	    pEntry ;
    char *	    pKey ;
    const char *    c ;
    int		    numTokens ;
    struct tToken * pTable ;
    struct tToken * p ;
    IV		    l	 ;
    int		    n ;
    int		    i ;
    unsigned char * pStartChars = pTokenTable -> cStartChars ;
    unsigned char * pAllChars	= pTokenTable -> cAllChars ;
    
    memset (pStartChars, 0, sizeof (pTokenTable -> cStartChars)) ;
    memset (pAllChars,   0, sizeof (pTokenTable -> cAllChars)) ;
    pTokenTable -> bLSearch = 0 ;    
    pTokenTable -> pContainsToken = NULL ;

    numTokens = 1 ;
    hv_iterinit (pTokenHash) ;
    while ((pEntry = hv_iternext (pTokenHash)))
	{
	pKey     = hv_iterkey (pEntry, &l) ;
	pToken   = hv_iterval (pTokenHash, pEntry) ;
        if (*pKey != '-') 
	    numTokens++ ;
        }
            
    if ((pTable = parse_malloc (r, sizeof (struct tToken) * numTokens)) == NULL)
         return rcOutOfMemory ;

    n = 0 ;
    hv_iterinit (pTokenHash) ;
    while ((pEntry = hv_iternext (pTokenHash)))
        {
        HV *   pHash ;
	struct tTokenTable * pNewTokenTable ;
	char *  sContains ;
        
        pKey     = hv_iterkey (pEntry, &l) ;
        pToken   = hv_iterval (pTokenHash, pEntry) ;
        
	if (*pKey == '-')
	    { /* special key */
	    if (strcmp (pKey, "-lsearch") == 0)
		{
		pTokenTable -> bLSearch = SvIV ((SV *)pToken) ;
		}
	    }
	else
	    {
	    if (!SvROK (pToken) || SvTYPE (SvRV (pToken)) != SVt_PVHV)
		{
		strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
		sprintf (r -> errdat2, "%s", pKey) ;
		return rcNotHashRef ;
		}
	    pHash = (HV *)SvRV (pToken) ;
        
	    p = &pTable[n] ;
	    p -> sName     = pKey ;
	    p -> sText     = GetHashValueStrDup (pHash, "text", "") ;
	    p -> nTextLen  = strlen (p -> sText) ;
	    p -> sEndText  = GetHashValueStrDup (pHash, "end", (char *)pDefEnd) ;
	    p -> sNodeName = GetHashValueStrDup (pHash, "nodename", NULL) ;
	    p -> nNodeType = GetHashValueInt (pHash, "nodetype", ntypTag) ;
	    p -> bUnescape = GetHashValueInt (pHash, "unescape", 0) ;
	    p -> nCDataType = GetHashValueInt (pHash, "cdatatype", ntypCDATA) ;
	    p -> nForceType = GetHashValueInt (pHash, "forcetype", 0) ;
	    p -> pStartTag  = (struct tToken *)GetHashValueStrDup (pHash, "starttag", NULL) ;
	    p -> pEndTag    = (struct tToken *)GetHashValueStrDup (pHash, "endtag", NULL) ;
	    if (sContains  = GetHashValueStrDup (pHash, "contains", NULL))
		{
		unsigned char * pC ;
		if ((p -> pContains = parse_malloc (r, sizeof (tCharMap))) == NULL)
		    return rcOutOfMemory ;

                pC = p -> pContains ;
		memset (pC, 0, sizeof (tCharMap)) ;
		while (*sContains)
		    {
		    pC[*sContains >> 3] |= 1 << (*sContains & 7) ;
		    sContains++ ;
		    }
		}
	    else
		p -> pContains = NULL ;

	    c = p -> sText ;
	    pStartChars [toupper(*c) >> 3] |= 1 << (toupper(*c) & 7) ;
	    pStartChars [tolower(*c) >> 3] |= 1 << (tolower(*c) & 7) ;
        
	    while (*c)
		{
		pAllChars [tolower(*c) >> 3] |= 1 << (tolower(*c) & 7) ;
		pAllChars [toupper(*c) >> 3] |= 1 << (toupper(*c) & 7) ;
		c++ ;
		}


	    if (r -> bDebug & dbgBuildToken)
                lprintf (r, "[%d]TOKEN: %s ... %s  unesc=%d nodetype=%d, cdatatype=%d, nodename=%s\n", r -> nPid, p -> sText, p -> pContains?sContains:p -> sEndText, p -> bUnescape, p -> nNodeType, p -> nCDataType, p -> sNodeName?p -> sNodeName:"<null>") ; 
        
	    if (p -> sNodeName)
		p -> nNodeName = String2Ndx (p -> sNodeName, strlen (p -> sNodeName)) ;
	    else
		p -> nNodeName = String2Ndx (p -> sText, strlen (p -> sText)) ;


	    if ((rc = CheckProcInfo (r, pHash, p)) != ok)
		return rc ;

	    
	    if ((rc = BuildSubTokenTable (r, pHash, pKey, "follow", p -> sEndText, &pNewTokenTable)))
		return rc ;
	    p -> pFollowedBy = pNewTokenTable ;

	    if ((rc = BuildSubTokenTable (r, pHash, pKey, "inside", "", &pNewTokenTable)))
		return rc ;
	    p -> pInside     = pNewTokenTable ;

	    n++ ;
	    }
	}

    qsort (pTable, numTokens - 1, sizeof (struct tToken), pTokenTable -> bLSearch?CmpTokenDesc:CmpToken) ;


    for (i = 0; i < n; i++)
	{
	if (pTable[i].pContains)
	    pTokenTable -> pContainsToken = &pTable[i] ;
        if (pTable[i].pEndTag)
	    {
	    char * s = (char *)pTable[i].pEndTag ;
	    int    j ;

	    pTable[i].pEndTag = NULL ;
	    for (j = 0; j < n; j++)
		{
		if (strcmp (pTable[j].sName, s) == 0)
		    pTable[i].pEndTag = &pTable[j] ;
		}
	    if (pTable[i].pEndTag == NULL)
		{
		strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
		sprintf (r -> errdat2, " EndTag %s for %s not found", pTable[i].sText, s) ;
		return rcNotFound ;
		}
	    
	    }
        if (pTable[i].pStartTag)
	    {
	    char * s = (char *)pTable[i].pStartTag ;
	    int    j ;

	    pTable[i].pStartTag = NULL ;
	    for (j = 0; j < n; j++)
		{
		if (strcmp (pTable[j].sName, s) == 0)
		    pTable[i].pStartTag = &pTable[j] ;
		}
	    if (pTable[i].pStartTag == NULL)
		{
		strncpy (r -> errdat1, "BuildTokenHash", sizeof (r -> errdat1)) ;
		sprintf (r -> errdat2, " StartTag %s for %s not found", pTable[i].sText, s) ;
		return rcNotFound ;
		}
	    
	    }
        }

    
    p = &pTable[n] ;
    p -> sText = "" ;
    p -> nTextLen = 0 ;
    p -> sEndText = "" ;
    p -> pFollowedBy = NULL ;
    p -> pInside     = NULL ;
    
        
    pTokenTable -> pTokens   = pTable ;
    pTokenTable -> numTokens = numTokens - 1 ;
            
    return ok ;
    }
    
    
/* ------------------------------------------------------------------------ */
/* compare tokens                                                           */
/* ------------------------------------------------------------------------ */

static int CmpTokenN (/*in*/ const void *  p1,
                     /*in*/ const void *  p2)

    {
    struct tTokenCmp * c = (struct tTokenCmp *)p1 ;
    int                i ;
    int	p1Len = c -> nLen ;
    int p2Len = ((struct tToken *)p2) -> nTextLen ;

    if ((i = strnicmp (c -> pStart, *((const char * *)p2), p1Len)) == 0)
	{
	if (p1Len == p2Len)
	    return 0 ;
	else if (p1Len > p2Len)
	    return 1 ;
	return -1 ;
	}
    return i ;
    }


/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ParseTokens                                                              */
/*                                                                          */
/* Parse a text for tokens                                                  */
/*                                                                          */
/* ------------------------------------------------------------------------ */

int ParseTokens (/*i/o*/ register req *	  r,
		 /*in */ char * *	  ppCurr, 
                    char *		  pEnd, 
                    struct tTokenTable *  pTokenTable, 
                    const char *          sEndText,
                    const char *          pParentContains,
		    enum tNodeType	  nCDataType,
		    enum tNodeType	  nForceType,
		    int			  bUnescape,
		    const char *          pParentNodeName,
		    tNode		  xParentNode,
		    int			  level) 

    {
    unsigned char * pStartChars = pTokenTable -> cStartChars ;
    struct tTokenCmp c ;    
    int nEndText = sEndText?strlen (sEndText):0 ;    
    char * pCurr = *ppCurr  ;
    char * pCurrStart = pCurr ;
    tNode xNewNode ;
    int	    rc ;
    tDomTree * pDomTree = DomTree_self (r -> xCurrDomTree) ;
    
    while (pCurr < pEnd)
        {
	struct tToken *	    pToken	    = NULL ;
        if (pStartChars [*pCurr >> 3] & 1 << (*pCurr & 7))
            { /* valid token start char found */
	    struct tTokenTable *    pNextTokenTab   = pTokenTable ;
	    const char *	    pNodeName	    = NULL ;
	    enum tNodeType	    nNodeType	    = 0 ;
	    char *		    pCurrTokenStart = pCurr ;

	    
	    do
                {
		struct tToken * pTokenTab = pNextTokenTab -> pTokens ;
		int             numTokens = pNextTokenTab -> numTokens ;
		unsigned char * pAllChars = pNextTokenTab -> cAllChars ;

		if (pNextTokenTab -> bLSearch)
		    { /* search linear thru the tokens */
		    int r = 1 ;
		    int i ;

		    for (i = 0, pToken = pTokenTab; i < numTokens; i++, pToken++)
			{
			r = strnicmp (pCurr, pToken -> sText, pToken -> nTextLen)  ;
			if (r == 0 || *pCurr > *(pToken -> sText))
			    break ;
			}
		    if (r != 0)
		    	pToken = NULL ;
		    else
			pCurr += pToken -> nTextLen ;
		    }
		else
		    { /* do a binary search for tokens */
		    c.pStart = pCurr ;

		    while (pAllChars [*pCurr >> 3] & (1 << (*pCurr & 7)))
			pCurr++ ;
                
		    c.nLen = pCurr - c.pStart ;
		    pToken = (struct tToken *)bsearch (&c, pTokenTab, numTokens, sizeof (struct tToken), CmpTokenN) ;
		    }

		if (pToken)
                    {
		    if (pToken -> nNodeType != ntypCDATA)
			while (isspace (*pCurr))
			    pCurr++ ;

                    if (pToken -> sNodeName)
			pNodeName = pToken -> sNodeName ;
		    }
                else
		    {
		    pToken = pNextTokenTab -> pContainsToken ;
		    break ;
		    }
		}
            while (pNextTokenTab = pToken -> pFollowedBy) ;
            
            if (pToken)
                { /* matching token found */       
                struct tTokenTable * pInside ;                

		if (pCurrStart < pCurrTokenStart)
		    {
		    if (nCDataType)
			{ /* add text before current token as node */
			if (bUnescape)
                            {
                            r -> bEscInUrl = bUnescape - 1 ;
                            TransHtml (r, pCurrStart, pCurrTokenStart - pCurrStart) ;
                            r -> bEscInUrl = 0 ;
                            }

			if (!(xNewNode = Node_appendChild (pDomTree, nCDataType, 0, pCurrStart, pCurrTokenStart - pCurrStart, xParentNode, level, GetLineNoOf (r, pCurrStart))))
			    return 1 ;
			}
		    pCurrStart = pCurrTokenStart ;
		    }
            
		if (pNodeName == NULL)
		    pNodeName = pToken -> sText ;
		
		if (pToken -> nNodeType == ntypEndTag && level > 0)
		    { /* end token found */
		    tNodeData * pStartTag ;
		    char * pEndCurr = strstr (pCurr, pToken -> sEndText) ;
                    if (pEndCurr)
			{ 
			tNode xNewAttrNode ;
                        if (pEndCurr - pCurr && pToken -> nCDataType && pToken -> nCDataType != ntypCDATA)
			    { /* add text before end of token as node */
			    if (pToken -> bUnescape)
                                {
                                r -> bEscInUrl = pToken -> bUnescape - 1 ;
				TransHtml (r, pCurr, pEndCurr - pCurr) ;
                                r -> bEscInUrl = 0 ;
                                }

			    if (!(xNewAttrNode = Node_appendChild (pDomTree, pToken -> nCDataType, 0, pCurr, pEndCurr - pCurr, xParentNode, level+1, GetLineNoOf (r, pCurr))))
				return 1 ;
			    }
			pCurr = pEndCurr + strlen (pToken -> sEndText) ;
			}
		    level-- ;
		    xParentNode = Node_parentNode  (pDomTree, xParentNode) ;
		    pStartTag = Node_selfLastChild (pDomTree, Node_self (pDomTree, xParentNode)) ;
		    if (pToken -> pStartTag == NULL || pStartTag -> nText != pToken -> pStartTag -> nNodeName)
			{
			strncpy (r -> errdat2, Ndx2String (pStartTag -> nText), sizeof (r -> errdat2)) ;	
			strncpy (r -> errdat1, Ndx2String (pToken -> nNodeName), sizeof (r -> errdat1)) ;	
			r -> Buf.pCurrPos = pCurrTokenStart ;
			return rcTagMismatch ;
			}
		    }
		else
		    {
		    if (pToken -> nNodeType == ntypEndStartTag && level > 0)
			{
			xParentNode = Node_parentNode  (pDomTree, xParentNode) ;
			level-- ;
			}
		    /* add token as node */
                    if (!(xNewNode = Node_appendChild (pDomTree, pToken -> nNodeType, (nCDataType == ntypAttrValue && pToken -> nNodeType != ntypAttr)?(pToken -> nForceType?2:1):0, pNodeName, strlen (pNodeName), xParentNode, level, GetLineNoOf (r, pCurrTokenStart))))
			{
			r -> Buf.pCurrPos = pCurrTokenStart ;

			return rc ;
			}
		    if (pInside = pToken -> pInside)
			{ /* parse for further tokens inside of this token */
                        ParseTokens (r, &pCurr, pEnd, pInside, pToken -> sEndText, pToken -> pContains, pToken -> nCDataType == ntypCDATA?ntypAttrValue:pToken -> nCDataType, 0, pToken -> bUnescape, pNodeName, xNewNode, level+1) ;
			}    
		    else
			{ /* nothing more inside of this token allowed, so search for the end of the token */
			char * pEndCurr ;
			unsigned char * pContains ;
			int nSkip ;
			if ((pContains = pToken -> pContains))
			    {
			    pEndCurr = pCurr ;
			    while (pContains [*pEndCurr >> 3] & (1 << (*pEndCurr & 7)))
                                pEndCurr++ ;
			    nSkip = 0 ;
			    }
			else
			    {
			    pEndCurr = strstr (pCurr, pToken -> sEndText) ;
			    nSkip = strlen (pToken -> sEndText) ;
			    }

			if (pEndCurr)
			    {
			    tNode xNewAttrNode ;
                            if (pEndCurr - pCurr && pToken -> nCDataType)
				{
				if (pToken -> bUnescape)
                                    {
                                    r -> bEscInUrl = pToken -> bUnescape - 1 ;
				    TransHtml (r, pCurr, pEndCurr - pCurr) ;
                                    r -> bEscInUrl = 0 ;
                                    }

				if (!(xNewAttrNode = Node_appendChild (pDomTree, pToken -> nCDataType, 0, pCurr, pEndCurr - pCurr, xNewNode, level+1, GetLineNoOf (r, pCurr))))
				    return 1 ;
				}
			    pCurr = pEndCurr + nSkip ;
			    }
			}

		    if (pToken -> nNodeType == ntypStartTag || pToken -> nNodeType == ntypEndStartTag)
			{
			level++ ;
			xParentNode = xNewNode ;
		        nCDataType = ntypCDATA ;
			}
		    }
		pCurrStart = pCurr ;
                }
	    }
        if (pParentContains && ((pParentContains [*pCurr >> 3] & 1 << (*pCurr & 7)) == 0) )
            {
	    if (pCurr - pCurrStart && nCDataType)
		{
		if (!(xNewNode = Node_appendChild (pDomTree, nCDataType, 0, pCurrStart, pCurr - pCurrStart, xParentNode, level, GetLineNoOf (r, pCurrStart))))
		    return 1 ;
		}
            *ppCurr = pCurr ;
            return 0 ;
            }
        else if (sEndText == NULL ||
	    (*pCurr == *sEndText && strncmp (pCurr, sEndText, nEndText) == 0))
            {
	    if ((pCurr - pCurrStart != 0 || nCDataType == ntypAttrValue) && nCDataType)
		if (!(xNewNode = Node_appendChild (pDomTree,  nCDataType, 0, pCurrStart, pCurr - pCurrStart, xParentNode, level, GetLineNoOf (r, pCurr))))
		    return 1 ;
            pCurr += nEndText ;
            *ppCurr = pCurr ;
            return 0 ;
            }
        else if (!pToken)
	    pCurr++ ;
        }
        
    if (nCDataType)
	if (!(xNewNode = Node_appendChild (pDomTree,  nCDataType, 0,pCurrStart, pCurr - pCurrStart, xParentNode, level, GetLineNoOf (r, pCurrStart))))
	    return 1 ;

    *ppCurr = pCurr ;
    return 0 ;
    }



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* ParseFile                                                                */
/*                                                                          */
/* Parse a source file                                                      */
/*                                                                          */
/* ------------------------------------------------------------------------ */


int ParseFile (/*i/o*/ register req * r)

    {
    char * pStart = r -> Buf.pBuf ;
    char * pEnd   = r -> Buf.pEndPos ;
    int	    rc ;
    tNode   xDocNode ;
    tNode   xNode ;
    tDomTree * pDomTree = DomTree_self (r -> xCurrDomTree) ;


    if (!(r -> xCurrDomTree  = DomTree_new (&pDomTree)))
	return 1 ;

    pDomTree -> xFilename = String2Ndx (r -> Buf.pFile -> sSourcefile, strlen (r -> Buf.pFile -> sSourcefile)) ;
    
    if (!(xDocNode = Node_appendChild (pDomTree,  ntypTag, 0, "attr", 3, 0, 0, 0)))
	return 1 ;

    if (!(xDocNode = Node_appendChild (pDomTree,  r -> bSubReq?ntypDocumentFraq:ntypDocument, 0, r -> bSubReq?"DocumentFraq":"Document", r -> bSubReq?12:8, 0, 0, 0)))
	return 1 ;
    
    if (!(xNode = Node_appendChild (pDomTree,  ntypAttr, 0, NULL, xDomTreeAttr, xDocNode, 0, 0)))
	return 1 ;

    if (!(xNode = Node_appendChild (pDomTree,  ntypAttrValue, 0, NULL, r -> xCurrDomTree, xNode, 0, 0)))
	return 1 ;

    pDomTree -> xDocument = xDocNode ;
    r -> xDocument = xDocNode ;

    if ((rc = ParseTokens (r, &pStart, pEnd, r -> pTokenTable, "", NULL, ntypCDATA, 0, 0, "root", xDocNode, 0)) != ok)
	return rc ; 



    return ok ;
    }



